import asyncio
from contextlib import (
    AbstractContextManager,
)
from contextlib import (
    nullcontext as does_not_raise,
)
from typing import (
    Any,
    AsyncIterator,
    ContextManager,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    cast,
)

import mock
import pytest
from decoy import Decoy

from opentrons_hardware.drivers.binary_usb import SerialUsbDriver
from opentrons_hardware.drivers.can_bus.abstract_driver import AbstractCanDriver
from opentrons_hardware.drivers.can_bus.can_messenger import (
    CanMessenger,
    MessageListenerCallback,
    MessageListenerCallbackFilter,
)
from opentrons_hardware.drivers.eeprom import EEPROMDriver
from opentrons_hardware.firmware_bindings.arbitration_id import ArbitrationId
from opentrons_hardware.firmware_bindings.constants import (
    NodeId,
    USBTarget,
)
from opentrons_hardware.firmware_bindings.constants import (
    PipetteName as FirmwarePipetteName,
)
from opentrons_hardware.firmware_bindings.messages.messages import MessageDefinition
from opentrons_hardware.firmware_bindings.utils import UInt8Field
from opentrons_hardware.hardware_control import current_settings
from opentrons_hardware.hardware_control.motion import (
    MoveGroupSingleAxisStep,
    MoveStopCondition,
    MoveType,
)
from opentrons_hardware.hardware_control.move_group_runner import MoveGroupRunner
from opentrons_hardware.hardware_control.network import DeviceInfoCache
from opentrons_hardware.hardware_control.tools.types import (
    GripperInformation,
    PipetteInformation,
    ToolSummary,
)
from opentrons_hardware.hardware_control.types import (
    MotorPositionStatus,
    MoveCompleteAck,
    PCBARevision,
)
from opentrons_shared_data.errors.exceptions import (
    EStopActivatedError,
    EStopNotPresentError,
    FailedGripperPickupError,
    FirmwareUpdateRequiredError,
    PipetteLiquidNotFoundError,
)

from opentrons.config.robot_configs import build_config_ot3
from opentrons.config.types import (
    GantryLoad,
    LiquidProbeSettings,
    OT3Config,
)
from opentrons.hardware_control.backends.estop_state import EstopStateMachine
from opentrons.hardware_control.backends.ot3controller import OT3Controller
from opentrons.hardware_control.backends.ot3utils import (
    axis_to_node,
    node_to_axis,
    sensor_node_for_mount,
    subsystem_to_target,
    target_to_subsystem,
)
from opentrons.hardware_control.backends.subsystem_manager import SubsystemManager
from opentrons.hardware_control.errors import (
    InvalidPipetteModel,
    InvalidPipetteName,
)
from opentrons.hardware_control.types import (
    Axis,
    CurrentConfig,
    EstopState,
    MotorStatus,
    OT3AxisMap,
    OT3Mount,
    SubSystem,
    SubSystemState,
    UpdateState,
    UpdateStatus,
)


@pytest.fixture
def mock_config() -> OT3Config:
    return build_config_ot3({})


class MockCanMessageNotifier:
    """A CanMessage notifier."""

    def __init__(self) -> None:
        """Constructor."""
        self._listeners: List[
            Tuple[MessageListenerCallback, Optional[MessageListenerCallbackFilter]]
        ] = []

    def add_listener(
        self,
        listener: MessageListenerCallback,
        filter: Optional[MessageListenerCallbackFilter] = None,
    ) -> None:
        """Add listener."""
        self._listeners.append((listener, filter))

    def notify(self, message: MessageDefinition, arbitration_id: ArbitrationId) -> None:
        """Notify."""
        for listener, filter in self._listeners:
            if filter and not filter(arbitration_id):
                continue
            listener(message, arbitration_id)


@pytest.fixture
def can_message_notifier() -> MockCanMessageNotifier:
    """A fixture that notifies mock_messenger listeners of a new message."""
    return MockCanMessageNotifier()


@pytest.fixture
def mock_messenger(can_message_notifier: MockCanMessageNotifier) -> mock.AsyncMock:
    """Mock can messenger."""
    messenger = mock.AsyncMock(spec=CanMessenger)
    messenger.add_listener.side_effect = can_message_notifier.add_listener
    return messenger


@pytest.fixture
def mock_can_driver(mock_messenger: mock.AsyncMock) -> AbstractCanDriver:
    driver = mock.AsyncMock(spec=AbstractCanDriver)

    # ignoring this type error because this is a very weird function that will in fact
    # do nothing, but has to have the yield in there for the compiler to make it a
    # generator function
    async def _fake_message_retrieve(d):  # type: ignore[no-untyped-def]
        while True:
            await asyncio.sleep(1)
        yield

    driver.__aiter__ = _fake_message_retrieve
    return driver


@pytest.fixture
def mock_usb_driver() -> SerialUsbDriver:
    return mock.AsyncMock(spec=SerialUsbDriver)


@pytest.fixture
def mock_eeprom_driver() -> EEPROMDriver:
    """Mock eeprom driver."""
    return mock.Mock(spec=EEPROMDriver)


@pytest.fixture
async def controller(
    mock_config: OT3Config,
    mock_can_driver: AbstractCanDriver,
    mock_eeprom_driver: EEPROMDriver,
) -> AsyncIterator[OT3Controller]:
    with mock.patch("opentrons.hardware_control.backends.ot3controller.OT3GPIO"):
        controller = OT3Controller(
            mock_config, mock_can_driver, eeprom_driver=mock_eeprom_driver
        )
        try:
            yield controller
        finally:
            await controller.clean_up()


@pytest.fixture
def fake_liquid_settings() -> LiquidProbeSettings:
    return LiquidProbeSettings(
        mount_speed=40,
        plunger_speed=10,
        plunger_impulse_time=0.2,
        sensor_threshold_pascals=15,
        aspirate_while_sensing=False,
        z_overlap_between_passes_mm=0.1,
        plunger_reset_offset=2.0,
        samples_for_baselining=20,
        sample_time_sec=0.004,
    )


@pytest.fixture
def mock_send_stop_threshold() -> Iterator[mock.AsyncMock]:
    with mock.patch(
        "opentrons_hardware.sensors.sensor_driver.SensorDriver.send_stop_threshold",
        autospec=True,
    ) as mock_stop_threshold:
        yield mock_stop_threshold


@pytest.fixture
def mock_move_group_run() -> Iterator[mock.AsyncMock]:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner.run",
        autospec=True,
    ) as mock_mgr_run:
        mock_mgr_run.return_value = {}
        yield mock_mgr_run


@pytest.fixture
def mock_check_overpressure() -> Iterator[mock.AsyncMock]:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.check_overpressure",
        autospec=True,
    ) as mock_check_overpressure:
        queue: asyncio.Queue[Any] = asyncio.Queue()

        class FakeOverpressure:
            async def __aenter__(self) -> asyncio.Queue[Any]:
                return queue

            async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
                pass

        mock_check_overpressure.return_value = lambda: FakeOverpressure()
        yield mock_check_overpressure


def _device_info_entry(subsystem: SubSystem) -> Tuple[SubSystem, DeviceInfoCache]:
    return subsystem, DeviceInfoCache(
        target=subsystem_to_target(subsystem),
        version=2,
        flags=0,
        shortsha="abcdef",
        revision=PCBARevision(main="c2"),
        subidentifier=1,
        ok=True,
    )


def _subsystems_entry(info: DeviceInfoCache) -> Tuple[SubSystem, SubSystemState]:
    return target_to_subsystem(info.target), SubSystemState(
        ok=info.ok,
        current_fw_version=info.version,
        next_fw_version=2,
        current_fw_sha=info.shortsha,
        pcba_revision="A1.0",
        update_state=None,
        fw_update_needed=False,
    )


@pytest.fixture
def mock_present_devices(
    controller: OT3Controller, mock_subsystem_manager: SubsystemManager, decoy: Decoy
) -> None:
    decoy.when(mock_subsystem_manager.device_info).then_return(
        dict(
            _device_info_entry(subsys)
            for subsys in (
                SubSystem.pipette_left,
                SubSystem.gantry_x,
                SubSystem.gantry_y,
                SubSystem.head,
                SubSystem.pipette_right,
                SubSystem.gripper,
                SubSystem.rear_panel,
            )
        )
    )
    decoy.when(mock_subsystem_manager.targets).then_return(
        {
            NodeId.pipette_left,
            NodeId.pipette_right,
            NodeId.gripper,
            NodeId.gantry_x,
            NodeId.gantry_y,
            NodeId.head,
            USBTarget.rear_panel,
        }
    )


@pytest.fixture
def mock_subsystem_manager(
    controller: OT3Controller, decoy: Decoy
) -> Iterator[SubsystemManager]:
    with mock.patch.object(
        controller, "_subsystem_manager", decoy.mock(cls=SubsystemManager)
    ) as mock_subsystem:
        yield mock_subsystem


@pytest.fixture
def mock_estop_state_machine(
    controller: OT3Controller, decoy: Decoy
) -> Iterator[EstopStateMachine]:
    with mock.patch.object(
        controller, "_estop_state_machine", decoy.mock(cls=EstopStateMachine)
    ) as mock_estop_state:
        yield mock_estop_state


@pytest.fixture
def fw_update_info() -> Dict[NodeId, str]:
    return {
        NodeId.head: "/some/path/head.hex",
        NodeId.gantry_x: "/some/path/gantry.hex",
    }


@pytest.fixture
def fw_node_info() -> Dict[NodeId, DeviceInfoCache]:
    node_cache1 = DeviceInfoCache(
        NodeId.head, 1, "12345678", None, PCBARevision(None), subidentifier=0, ok=True
    )
    node_cache2 = DeviceInfoCache(
        NodeId.gantry_x,
        1,
        "12345678",
        None,
        PCBARevision(None),
        subidentifier=0,
        ok=True,
    )
    return {NodeId.head: node_cache1, NodeId.gantry_x: node_cache2}


home_test_params = [
    [Axis.X],
    [Axis.Y],
    [Axis.Z_L],
    [Axis.Z_R],
    [Axis.X, Axis.Y, Axis.Z_R],
    [Axis.X, Axis.Z_R, Axis.P_R, Axis.Y, Axis.Z_L],
    [Axis.X, Axis.Y, Axis.Z_L, Axis.Z_R, Axis.P_L, Axis.P_R],
    [Axis.P_R],
    [Axis.Z_L, Axis.Z_R, Axis.Z_G],
    [Axis.X, Axis.Y, Axis.Z_G],
]


def move_group_run_side_effect_home(
    controller: OT3Controller, axes_to_home: List[Axis]
) -> Iterator[Dict[NodeId, MotorPositionStatus]]:
    """Return homed position for axis that is present and was commanded to home."""
    motor_nodes = controller._motor_nodes()
    gantry_homes = {
        axis_to_node(ax): MotorPositionStatus(0.0, 0.0, True, True, MoveCompleteAck(1))
        for ax in Axis.gantry_axes()
        if ax in axes_to_home and axis_to_node(ax) in motor_nodes
    }
    if gantry_homes:
        yield gantry_homes

    pipette_homes = {
        axis_to_node(ax): MotorPositionStatus(0.0, 0.0, True, True, MoveCompleteAck(1))
        for ax in Axis.pipette_axes()
        if ax in axes_to_home and axis_to_node(ax) in motor_nodes
    }
    yield pipette_homes


@pytest.mark.parametrize("axes", home_test_params)
async def test_home_execute(
    controller: OT3Controller,
    axes: List[Axis],
    mock_present_devices: None,
    mock_check_overpressure: None,
) -> None:
    config = {"run.side_effect": move_group_run_side_effect_home(controller, axes)}
    with mock.patch(  # type: ignore [call-overload]
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner",
        spec=MoveGroupRunner,
        **config,
    ) as mock_runner:
        present_axes = set(ax for ax in axes if controller.axis_is_present(ax))
        controller.set_pressure_sensor_available(Axis.P_L, True)
        controller.set_pressure_sensor_available(Axis.P_R, True)

        # nothing has been homed
        assert not controller._motor_status
        await controller.home(axes, GantryLoad.LOW_THROUGHPUT)
        all_groups = [
            group
            for arg in mock_runner.call_args_list
            for group in arg.kwargs["move_groups"]
        ]

        actual_nodes_steps: Dict[Axis, List[MoveGroupSingleAxisStep]] = {
            ax: [] for ax in axes
        }
        for group in all_groups:
            for step in group:
                for k, v in step.items():
                    actual_nodes_steps[node_to_axis(k)].append(v)

        # every single node will receive one home request and one backoff requests
        for ax in present_axes:
            assert len(actual_nodes_steps[ax]) == 2
            home_request = filter(
                lambda m: m.stop_condition == MoveStopCondition.limit_switch,
                actual_nodes_steps[ax],
            )
            backoff_request = filter(
                lambda m: m.stop_condition == MoveStopCondition.limit_switch_backoff,
                actual_nodes_steps[ax],
            )
            assert len(list(home_request)) == 1
            assert len(list(backoff_request)) == 1

        # all commanded axes have been homed
        assert all(controller._motor_status[axis_to_node(ax)].motor_ok for ax in axes)
        assert controller.check_motor_status(axes)


@pytest.mark.parametrize("axes", home_test_params)
async def test_home_gantry_order(
    controller: OT3Controller,
    mock_move_group_run: mock.AsyncMock,
    axes: List[Axis],
    mock_present_devices: None,
) -> None:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner",
        spec=MoveGroupRunner,
    ) as mock_runner:
        controller._build_home_gantry_z_runner(axes, GantryLoad.LOW_THROUGHPUT)
        has_mount = len(set(Axis.ot3_mount_axes()) & set(axes)) > 0
        has_x = Axis.X in axes
        has_y = Axis.Y in axes
        if has_mount or has_x or has_y:
            gantry_moves = mock_runner.call_args_list[0].kwargs["move_groups"]

            # mount steps are commanded first
            if has_mount:
                # only one seq per group
                assert len(gantry_moves[0]) == len(gantry_moves[1]) == 1
                assert gantry_moves[0][0].keys() == gantry_moves[1][0].keys()
                assert all(
                    node_to_axis(node) in Axis.ot3_mount_axes()
                    for node in gantry_moves[0][0].keys()
                )
                gantry_moves.pop(0)
                gantry_moves.pop(0)

            # then X
            if has_x:
                # only one seq per group
                assert len(gantry_moves[0]) == len(gantry_moves[1]) == 1
                assert gantry_moves[0][0].keys() == gantry_moves[1][0].keys()
                assert all(
                    node_to_axis(node) == Axis.X for node in gantry_moves[0][0].keys()
                )
                gantry_moves.pop(0)
                gantry_moves.pop(0)

            # lastly Y
            if has_y:
                # only one seq per group
                assert len(gantry_moves[0]) == len(gantry_moves[1]) == 1
                assert gantry_moves[0][0].keys() == gantry_moves[1][0].keys()
                assert all(
                    node_to_axis(node) == Axis.Y for node in gantry_moves[0][0].keys()
                )
                gantry_moves.pop(0)
                gantry_moves.pop(0)

            assert not gantry_moves


@pytest.mark.parametrize("axes", home_test_params)
async def test_home_only_present_devices(
    controller: OT3Controller,
    mock_move_group_run: mock.AsyncMock,
    axes: List[Axis],
    mock_present_devices: None,
    mock_check_overpressure: None,
) -> None:
    starting_position = {
        NodeId.head_l: 20.0,
        NodeId.head_r: 85.0,
        NodeId.gantry_x: 68.0,
        NodeId.gantry_y: 54.0,
        NodeId.pipette_left: 30.0,
        NodeId.pipette_right: 110.0,
    }
    homed_position = {}

    controller._position = starting_position
    controller.set_pressure_sensor_available(Axis.P_L, True)
    controller.set_pressure_sensor_available(Axis.P_R, True)

    mock_move_group_run.side_effect = move_group_run_side_effect_home(controller, axes)

    # nothing has been homed
    assert not controller._motor_status
    await controller.home(axes, GantryLoad.LOW_THROUGHPUT)

    for call in mock_move_group_run.call_args_list:
        # pull the bound-self argument that is the runner instance out of
        # the args list - we can do this because the mock here is the
        # function definition in the class body
        move_group_runner = call[0][0]
        for move_group in move_group_runner._move_groups:
            assert move_group  # don't pass in empty groups
            for move_group_step in move_group:
                assert move_group_step  # don't pass in empty moves
                for node, step in move_group_step.items():
                    assert step  # don't pass in empty steps
                    homed_position[node] = 0.0  # track homed position for node

    # check that the current position is updated
    expected_position = {**starting_position, **homed_position}
    for node, pos in controller._position.items():
        assert pos == expected_position[node]
    # check that the homed axis is tracked by _homed_nodes
    assert controller._motor_status.keys() == homed_position.keys()


async def test_get_attached_instruments(
    controller: OT3Controller, mock_subsystem_manager: SubsystemManager, decoy: Decoy
) -> None:
    pipette_id = "P1KSV33hello"
    gripper_id = "GRPV00fake_serial"
    gripper_name = "Flex Gripper"
    decoy.when(mock_subsystem_manager.tools).then_return(
        ToolSummary(
            left=PipetteInformation(
                name=FirmwarePipetteName.p1000_single,
                name_int=FirmwarePipetteName.p1000_single.value,
                model="3.3",
                serial="hello",
            ),
            right=None,
            gripper=GripperInformation(model="0.0", serial="fake_serial"),
        )
    )

    detected = await controller.get_attached_instruments({})
    assert list(detected.keys()) == [OT3Mount.LEFT, OT3Mount.GRIPPER]
    assert detected[OT3Mount.LEFT]["id"] == pipette_id
    gripper_obj = detected[OT3Mount.GRIPPER]
    assert gripper_obj
    assert gripper_obj["id"] == gripper_id
    config = gripper_obj["config"]
    assert config
    assert config.display_name == gripper_name


async def test_get_attached_instruments_handles_unknown_name(
    controller: OT3Controller, mock_subsystem_manager: SubsystemManager, decoy: Decoy
) -> None:
    decoy.when(mock_subsystem_manager.tools).then_return(
        ToolSummary(
            left=PipetteInformation(
                name=FirmwarePipetteName.unknown,
                name_int=41,
                model="30",
                serial="hello",
            ),
            right=None,
            gripper=GripperInformation(
                model="0.0",
                serial="fake_serial",
            ),
        )
    )
    with pytest.raises(InvalidPipetteName):
        await controller.get_attached_instruments({})


async def test_get_attached_instruments_handles_unknown_model(
    controller: OT3Controller, mock_subsystem_manager: SubsystemManager, decoy: Decoy
) -> None:
    decoy.when(mock_subsystem_manager.tools).then_return(
        ToolSummary(
            left=PipetteInformation(
                name=FirmwarePipetteName.p1000_single,
                name_int=0,
                model="4.1",
                serial="hello",
            ),
            right=None,
            gripper=GripperInformation(model="0", serial="fake_serial"),
        )
    )
    with pytest.raises(InvalidPipetteModel):
        await controller.get_attached_instruments({})


async def test_gripper_home_jaw(
    controller: OT3Controller, mock_move_group_run: mock.AsyncMock
) -> None:
    await controller.gripper_home_jaw(25)
    for call in mock_move_group_run.call_args_list:
        move_group_runner = call[0][0]
        for move_group in move_group_runner._move_groups:
            assert move_group  # don't pass in empty groups
            assert len(move_group) == 1
        # only homing the gripper jaw
        assert list(move_group[0].keys()) == [NodeId.gripper_g]
        step = move_group[0][NodeId.gripper_g]
        assert step.stop_condition == MoveStopCondition.limit_switch
        assert step.move_type == MoveType.home


async def test_gripper_grip(
    controller: OT3Controller, mock_move_group_run: mock.AsyncMock
) -> None:
    await controller.gripper_grip_jaw(duty_cycle=50, expected_displacement=0)
    for call in mock_move_group_run.call_args_list:
        move_group_runner = call[0][0]
        for move_group in move_group_runner._move_groups:
            assert move_group  # don't pass in empty groups
            assert len(move_group) == 1
        # only gripping the gripper jaw
        assert list(move_group[0].keys()) == [NodeId.gripper_g]
        step = move_group[0][NodeId.gripper_g]
        assert step.stop_condition == MoveStopCondition.none
        assert step.move_type == MoveType.grip


async def test_gripper_jaw_width(
    controller: OT3Controller, mock_move_group_run: mock.AsyncMock
) -> None:
    max_jaw_width = 134350
    await controller.gripper_hold_jaw(
        encoder_position_um=int((max_jaw_width - 80000) / 2)
    )
    for call in mock_move_group_run.call_args_list:
        move_group_runner = call[0][0]
        for move_group in move_group_runner._move_groups:
            assert move_group  # don't pass in empty groups
            assert len(move_group) == 1
        # only moving the gripper jaw
        assert list(move_group[0].keys()) == [NodeId.gripper_g]
        step = move_group[0][NodeId.gripper_g]
        assert step.stop_condition == MoveStopCondition.encoder_position
        assert step.move_type == MoveType.linear


async def test_get_limit_switches(
    controller: OT3Controller,
    mock_subsystem_manager: SubsystemManager,
    decoy: Decoy,
) -> None:
    decoy.when(mock_subsystem_manager.device_info).then_return(
        dict(
            (
                _device_info_entry(SubSystem.gantry_x),
                _device_info_entry(SubSystem.gantry_y),
            )
        )
    )
    decoy.when(mock_subsystem_manager.targets).then_return(
        {NodeId.gantry_x, NodeId.gantry_y}
    )

    fake_response = {
        NodeId.gantry_x: UInt8Field(0),
        NodeId.gantry_y: UInt8Field(0),
    }

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.get_limit_switches",
        autospec=True,
    ) as mock_hardware_get_limit_switches:
        mock_hardware_get_limit_switches.return_value = fake_response
        res = await controller.get_limit_switches()
        mock_hardware_get_limit_switches.assert_called_once_with(
            controller._messenger, {NodeId.gantry_x, NodeId.gantry_y}
        )
        assert Axis.X in res
        assert Axis.Y in res


@pytest.mark.parametrize(
    "motor_status,ready",
    [
        ({}, False),
        ({NodeId.gripper_g: MotorStatus(True, True)}, False),
        (
            {
                NodeId.gantry_x: MotorStatus(True, True),
                NodeId.gantry_y: MotorStatus(True, True),
                NodeId.head_l: MotorStatus(False, True),
            },
            False,
        ),
        (
            {
                NodeId.gantry_x: MotorStatus(True, True),
                NodeId.gantry_y: MotorStatus(True, True),
                NodeId.head_l: MotorStatus(True, True),
            },
            True,
        ),
    ],
)
async def test_ready_for_movement(
    controller: OT3Controller,
    motor_status: Dict[NodeId, MotorStatus],
    ready: bool,
) -> None:
    controller._motor_status = motor_status

    axes = [Axis.X, Axis.Y, Axis.Z_L]
    assert controller.check_motor_status(axes) == ready


def probe_move_group_run_side_effect(
    head: NodeId, tool: NodeId
) -> Iterator[Dict[NodeId, MotorPositionStatus]]:
    """Return homed position for axis that is present and was commanded to home."""
    positions = {
        head: MotorPositionStatus(0.0, 0.0, True, True, MoveCompleteAck(1)),
        tool: MotorPositionStatus(0.0, 0.0, True, True, MoveCompleteAck(1)),
    }
    yield positions


@pytest.mark.parametrize("mount", [OT3Mount.LEFT, OT3Mount.RIGHT])
async def test_liquid_probe(
    mount: OT3Mount,
    controller: OT3Controller,
    fake_liquid_settings: LiquidProbeSettings,
    mock_move_group_run: mock.AsyncMock,
    mock_send_stop_threshold: mock.AsyncMock,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    fake_max_p_dist = 70
    head_node = axis_to_node(Axis.by_mount(mount))
    tool_node = sensor_node_for_mount(mount)
    mock_move_group_run.side_effect = probe_move_group_run_side_effect(
        head_node, tool_node
    )
    controller._pipettes_to_monitor_pressure = mock.MagicMock(  # type: ignore[method-assign]
        return_value=[sensor_node_for_mount(mount)]
    )
    monkeypatch.setattr(
        "opentrons_hardware.hardware_control.tool_sensors.finalize_logs",
        mock.AsyncMock(),
    )
    try:
        await controller.liquid_probe(
            mount=mount,
            max_p_distance=fake_max_p_dist,
            mount_speed=fake_liquid_settings.mount_speed,
            plunger_speed=fake_liquid_settings.plunger_speed,
            threshold_pascals=fake_liquid_settings.sensor_threshold_pascals,
            plunger_impulse_time=fake_liquid_settings.plunger_impulse_time,
            num_baseline_reads=fake_liquid_settings.samples_for_baselining,
            z_offset_for_plunger_prep=2.0,
        )
    except PipetteLiquidNotFoundError:
        # the move raises a liquid not found now since we don't call the move group and it doesn't
        # get any positions back
        pass
    move_groups = mock_move_group_run.call_args_list[0][0][0]._move_groups
    #  in tool_sensors, pipette moves down, then sensor move goes
    assert move_groups[0][0][tool_node].stop_condition == MoveStopCondition.none
    assert (
        move_groups[1][0][tool_node].stop_condition == MoveStopCondition.sensor_report
    )
    assert len(move_groups) == 2
    assert move_groups[0][0][tool_node]
    assert move_groups[1][0][head_node], move_groups[2][0][tool_node]


async def test_tip_action(
    controller: OT3Controller,
    mock_move_group_run: mock.AsyncMock,
) -> None:
    await controller.home_tip_motors(distance=33, velocity=-5.5, back_off=False)
    for call in mock_move_group_run.call_args_list:
        move_group_runner = call[0][0]
        for move_group in move_group_runner._move_groups:
            assert move_group  # don't pass in empty groups
            assert len(move_group) == 1
        # we should be sending this command to the pipette axes to process.
        assert list(move_group[0].keys()) == [NodeId.pipette_left]
        step = move_group[0][NodeId.pipette_left]
        assert step.stop_condition == MoveStopCondition.limit_switch

    mock_move_group_run.reset_mock()

    await controller.home_tip_motors(distance=33, velocity=-5.5, back_off=True)
    for call in mock_move_group_run.call_args_list:
        move_group_runner = call[0][0]
        move_groups = move_group_runner._move_groups

        for move_group in move_groups:
            assert move_group  # don't pass in empty groups
            assert len(move_group) >= 1

        # we should be sending this command to the pipette axes to process.
        home_step = move_groups[0][0][NodeId.pipette_left]
        assert home_step.stop_condition == MoveStopCondition.limit_switch
        backoff_step = move_groups[0][1][NodeId.pipette_left]
        assert backoff_step.stop_condition == MoveStopCondition.limit_switch_backoff


async def test_update_motor_status(
    mock_messenger: CanMessenger, controller: OT3Controller, mock_present_devices: None
) -> None:
    async def fake_gmp(
        can_messenger: CanMessenger, nodes: Set[NodeId], timeout: float = 1.0
    ) -> Dict[NodeId, MotorPositionStatus]:
        return {
            node: MotorPositionStatus(0.223, 0.323, False, True, None) for node in nodes
        }

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.get_motor_position", fake_gmp
    ):
        nodes = set([NodeId.gantry_x, NodeId.gantry_y, NodeId.head_l, NodeId.head_r])
        await controller.update_motor_status()
        for node in nodes:
            assert controller._position.get(node) == 0.223
            assert controller._encoder_position.get(node) == 0.323
            assert controller._motor_status.get(node) == MotorStatus(False, True)


@pytest.mark.parametrize("axes", home_test_params)
async def test_update_motor_estimation(
    mock_messenger: CanMessenger,
    controller: OT3Controller,
    axes: List[Axis],
    mock_present_devices: None,
) -> None:
    async def fake_umpe(
        can_messenger: CanMessenger, nodes: Set[NodeId], timeout: float = 1.0
    ) -> Dict[NodeId, MotorPositionStatus]:
        return {
            node: MotorPositionStatus(0.223, 0.323, False, True, None) for node in nodes
        }

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.update_motor_position_estimation",
        fake_umpe,
    ):
        nodes = [axis_to_node(a) for a in axes]
        for node in nodes:
            controller._motor_status.update(
                {node: MotorStatus(motor_ok=False, encoder_ok=True)}
            )
        await controller.update_motor_estimation(axes)
        for node in nodes:
            assert controller._position.get(node) == 0.223
            assert controller._encoder_position.get(node) == 0.323
            assert controller._motor_status.get(node) == MotorStatus(False, True)


@pytest.mark.parametrize(
    argnames=["gantry_load", "expected_call"],
    argvalues=[
        [
            GantryLoad.HIGH_THROUGHPUT_1000,
            [NodeId.pipette_left],
        ],  # this uses the Q motor
        [GantryLoad.LOW_THROUGHPUT, []],
    ],
)
async def test_set_default_currents(
    controller: OT3Controller,
    gantry_load: GantryLoad,
    expected_call: List[NodeId],
    mock_present_devices: None,
) -> None:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ) as mocked_currents:
        await controller.update_to_default_current_settings(gantry_load)
        mocked_currents.assert_called_once_with(
            mocked_currents.call_args_list[0][0][0],
            mocked_currents.call_args_list[0][0][1],
            use_tip_motor_message_for=expected_call,
        )
        these_current_settings = controller._current_settings
        assert these_current_settings
        for k, v in these_current_settings.items():
            if k == Axis.P_L and (
                gantry_load == GantryLoad.HIGH_THROUGHPUT_1000
                and expected_call[0] == NodeId.pipette_left
            ):
                # q motor config
                v = these_current_settings[Axis.Q]
                assert (
                    mocked_currents.call_args_list[0][0][1][axis_to_node(k)]
                    == v.as_tuple()
                )
            else:
                assert (
                    mocked_currents.call_args_list[0][0][1][axis_to_node(k)]
                    == v.as_tuple()
                )


@pytest.mark.parametrize(
    argnames=["active_current", "gantry_load", "expected_call"],
    argvalues=[
        [
            {Axis.X: 1.0, Axis.Y: 2.0},
            GantryLoad.LOW_THROUGHPUT,
            [{NodeId.gantry_x: 1.0, NodeId.gantry_y: 2.0}, []],
        ],
        [
            {Axis.Q: 1.5},
            GantryLoad.HIGH_THROUGHPUT_1000,
            [{NodeId.pipette_left: 1.5}, [NodeId.pipette_left]],
        ],
    ],
)
async def test_set_run_current(
    controller: OT3Controller,
    active_current: OT3AxisMap[float],
    gantry_load: GantryLoad,
    expected_call: List[Any],
    mock_present_devices: None,
) -> None:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ):
        with mock.patch(
            "opentrons.hardware_control.backends.ot3controller.set_run_current",
            spec=current_settings.set_run_current,
        ) as mocked_currents:
            await controller.update_to_default_current_settings(gantry_load)
            await controller.set_active_current(active_current)
            mocked_currents.assert_called_once_with(
                mocked_currents.call_args_list[0][0][0],
                expected_call[0],
                use_tip_motor_message_for=expected_call[1],
            )


@pytest.mark.parametrize(
    argnames=["hold_current", "gantry_load", "expected_call"],
    argvalues=[
        [
            {Axis.P_L: 0.5, Axis.Y: 0.8},
            GantryLoad.LOW_THROUGHPUT,
            [{NodeId.pipette_left: 0.5, NodeId.gantry_y: 0.8}, []],
        ],
        [
            {Axis.Q: 0.8},
            GantryLoad.HIGH_THROUGHPUT_1000,
            [{NodeId.pipette_left: 0.8}, [NodeId.pipette_left]],
        ],
    ],
)
async def test_set_hold_current(
    controller: OT3Controller,
    hold_current: OT3AxisMap[float],
    gantry_load: GantryLoad,
    expected_call: List[Any],
    mock_present_devices: None,
) -> None:
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ):
        with mock.patch(
            "opentrons.hardware_control.backends.ot3controller.set_hold_current",
            spec=current_settings.set_hold_current,
        ) as mocked_currents:
            await controller.update_to_default_current_settings(gantry_load)
            await controller.set_hold_current(hold_current)
            mocked_currents.assert_called_once_with(
                mocked_currents.call_args_list[0][0][0],
                expected_call[0],
                use_tip_motor_message_for=expected_call[1],
            )


async def test_update_required_flag(
    mock_messenger: CanMessenger,
    controller: OT3Controller,
    mock_subsystem_manager: SubsystemManager,
    decoy: Decoy,
) -> None:
    """Test that FirmwareUpdateRequired is raised when update_required flag is set."""
    axes = [Axis.X, Axis.Y]
    decoy.when(mock_subsystem_manager.update_required).then_return(True)
    controller._initialized = True
    with pytest.raises(FirmwareUpdateRequiredError):
        await controller.home(axes, gantry_load=GantryLoad.LOW_THROUGHPUT)


async def test_update_required_bypass_firmware_update(
    controller: OT3Controller, mock_subsystem_manager: SubsystemManager, decoy: Decoy
) -> None:
    """Do not raise FirmwareUpdateRequired for update_firmware."""
    decoy.when(mock_subsystem_manager.update_required).then_return(True)

    async def _mock_update() -> AsyncIterator[UpdateStatus]:
        yield UpdateStatus(
            subsystem=SubSystem.gantry_x, state=UpdateState.done, progress=100
        )

    decoy.when(mock_subsystem_manager.update_firmware(set(), False)).then_return(
        _mock_update()
    )
    controller._initialized = True
    async for status_element in controller.update_firmware(set()):
        pass
    # raise FirmwareUpdateRequired if the _update_required flag is set
    controller._initialized = True
    with pytest.raises(FirmwareUpdateRequiredError):
        await controller.home([Axis.X], gantry_load=GantryLoad.LOW_THROUGHPUT)


async def test_update_required_flag_false(
    controller: OT3Controller,
    mock_subsystem_manager: SubsystemManager,
    decoy: Decoy,
    mock_present_devices: None,
) -> None:
    """Do not raise FirmwareUpdateRequired if update_required is False."""

    decoy.when(mock_subsystem_manager.device_info).then_return(
        dict(
            (
                _device_info_entry(SubSystem.gantry_x),
                _device_info_entry(SubSystem.gantry_y),
            )
        )
    )
    decoy.when(mock_subsystem_manager.update_required).then_return(False)

    for node in controller._motor_nodes():
        controller._motor_status.update(
            {node: MotorStatus(motor_ok=False, encoder_ok=True)}
        )
    # update_required is false so dont raise FirmwareUpdateRequired
    controller._initialized = True
    controller._check_updates = True
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ):
        await controller.update_to_default_current_settings(GantryLoad.LOW_THROUGHPUT)

    async def fake_src(
        can_messenger: CanMessenger,
        current_settings: Dict[NodeId, float],
        use_tip_motor_message_for: List[NodeId],
    ) -> None:
        return None

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_run_current",
        fake_src,
    ):
        await controller.set_active_current({Axis.X: 2})


async def test_update_required_flag_initialized(
    controller: OT3Controller,
    mock_subsystem_manager: SubsystemManager,
    mock_present_devices: None,
    decoy: Decoy,
) -> None:
    """Do not raise FirmwareUpdateRequired if initialized is False."""
    decoy.when(mock_subsystem_manager.update_required).then_return(True)

    for node in controller._motor_nodes():
        controller._motor_status.update(
            {node: MotorStatus(motor_ok=False, encoder_ok=True)}
        )

    # update_required is true, but initialized is false so dont raise FirmwareUpdateRequired
    controller._initialized = False
    controller._check_updates = True
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ):
        await controller.update_to_default_current_settings(GantryLoad.LOW_THROUGHPUT)

    async def fake_src(
        can_messenger: CanMessenger,
        current_settings: Dict[NodeId, float],
        use_tip_motor_message_for: List[NodeId],
    ) -> None:
        return None

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_run_current",
        fake_src,
    ):
        await controller.set_active_current({Axis.X: 2})


async def test_update_required_flag_disabled(
    controller: OT3Controller,
    mock_present_devices: None,
    decoy: Decoy,
    mock_subsystem_manager: SubsystemManager,
) -> None:
    """Do not raise FirmwareUpdateRequired if check_updates is False."""
    decoy.when(mock_subsystem_manager.update_required).then_return(True)

    for node in controller._motor_nodes():
        controller._motor_status.update(
            {node: MotorStatus(motor_ok=False, encoder_ok=True)}
        )
    # update_required and initialized are true, but not check_updates, no exception
    controller._initialized = False
    controller._check_updates = False
    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_currents",
        spec=current_settings.set_currents,
    ):
        await controller.update_to_default_current_settings(GantryLoad.LOW_THROUGHPUT)

    async def fake_src(
        can_messenger: CanMessenger,
        current_settings: Dict[NodeId, float],
        use_tip_motor_message_for: List[NodeId],
    ) -> None:
        return None

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_run_current",
        fake_src,
    ):
        await controller.set_active_current({Axis.X: 2})


@pytest.mark.parametrize(
    "estop_state, expectation",
    [
        [EstopState.DISENGAGED, does_not_raise()],
        [EstopState.NOT_PRESENT, pytest.raises(EStopNotPresentError)],
        [EstopState.PHYSICALLY_ENGAGED, pytest.raises(EStopActivatedError)],
        [EstopState.LOGICALLY_ENGAGED, pytest.raises(EStopActivatedError)],
    ],
)
async def test_requires_estop(
    controller: OT3Controller,
    mock_estop_state_machine: EstopStateMachine,
    decoy: Decoy,
    estop_state: EstopState,
    expectation: ContextManager[None],
    mock_check_overpressure: None,
) -> None:
    """Test that the estop state machine raises properly."""
    decoy.when(mock_estop_state_machine.state).then_return(estop_state)

    with expectation:
        await controller.home([Axis.X, Axis.Y], gantry_load=GantryLoad.LOW_THROUGHPUT)


@pytest.mark.parametrize(
    "run_currents, hold_currents",
    [
        [{Axis.X: 1.0}, {}],
        [{}, {Axis.X: 1.0}],
        [{Axis.X: 1.0}, {Axis.X: 1.0}],
        [{}, {}],
    ],
)
async def test_motor_current(
    controller: OT3Controller,
    run_currents: OT3AxisMap[float],
    hold_currents: OT3AxisMap[float],
) -> None:
    """Test that restore current actually works."""
    controller._current_settings = {Axis.X: CurrentConfig(0.0, 0.0)}

    with mock.patch.object(controller, "set_active_current") as mock_run_currents:
        with mock.patch.object(controller, "set_hold_current") as mock_hold_currents:
            with mock.patch.object(controller, "set_default_currents") as mock_default:
                async with controller.motor_current(run_currents, hold_currents):
                    await controller.update_position()

                if not run_currents and not hold_currents:
                    mock_default.assert_called_once()
                    mock_run_currents.assert_not_called()
                    mock_hold_currents.assert_not_called()
                elif run_currents:
                    mock_run_currents.assert_has_calls(
                        [
                            mock.call({Axis.X: 1.0}),
                            mock.call({Axis.X: 0.0}),
                        ],
                    )
                elif hold_currents:
                    mock_hold_currents.assert_has_calls(
                        [
                            mock.call({Axis.X: 1.0}),
                            mock.call({Axis.X: 0.0}),
                        ],
                    )


@pytest.mark.parametrize(
    ["axes", "expected_tip_nodes", "expected_normal_nodes"],
    [
        [
            [Axis.X, Axis.Z_L, Axis.Q],
            {NodeId.pipette_left},
            {NodeId.gantry_x, NodeId.head_l},
        ],
        [[Axis.X, Axis.Z_L], {}, {NodeId.gantry_x, NodeId.head_l}],
        [[Axis.Q], {NodeId.pipette_left}, {}],
        [
            [Axis.X, Axis.Z_L, Axis.P_L, Axis.Q],
            {NodeId.pipette_left},
            {NodeId.gantry_x, NodeId.head_l, NodeId.pipette_left},
        ],
    ],
)
async def test_engage_motors(
    controller: OT3Controller,
    axes: List[Axis],
    expected_tip_nodes: Set[NodeId],
    expected_normal_nodes: Set[NodeId],
) -> None:
    """Test that engaging/disengaging motors works."""

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_enable_motor",
        autospec=True,
    ) as set_normal_axes:
        with mock.patch(
            "opentrons.hardware_control.backends.ot3controller.set_enable_tip_motor",
            autospec=True,
        ) as set_tip_axes:
            await controller.engage_axes(axes=axes)

            if len(expected_normal_nodes) > 0:
                set_normal_axes.assert_awaited_with(
                    controller._messenger, expected_normal_nodes
                )
            else:
                set_normal_axes.assert_not_awaited()
            if len(expected_tip_nodes) > 0:
                set_tip_axes.assert_awaited_with(
                    controller._messenger, expected_tip_nodes
                )
            else:
                set_tip_axes.assert_not_awaited()

    with mock.patch(
        "opentrons.hardware_control.backends.ot3controller.set_disable_motor",
        autospec=True,
    ) as set_normal_axes:
        with mock.patch(
            "opentrons.hardware_control.backends.ot3controller.set_disable_tip_motor",
            autospec=True,
        ) as set_tip_axes:
            await controller.disengage_axes(axes=axes)

            if len(expected_normal_nodes) > 0:
                set_normal_axes.assert_awaited_with(
                    controller._messenger, expected_normal_nodes
                )
            else:
                set_normal_axes.assert_not_awaited()
            if len(expected_tip_nodes) > 0:
                set_tip_axes.assert_awaited_with(
                    controller._messenger, expected_tip_nodes
                )
            else:
                set_tip_axes.assert_not_awaited()


@pytest.mark.parametrize(
    "expected_grip_width,actual_grip_width,wider,narrower,allowed_error,hard_max,hard_min,raise_error",
    [
        (80, 80, 0, 0, 0, 92, 60, False),
        (80, 81, 0, 0, 0, 92, 60, True),
        (80, 79, 0, 0, 0, 92, 60, True),
        (80, 81, 1, 0, 0, 92, 60, False),
        (80, 79, 0, 1, 0, 92, 60, False),
        (80, 81, 0, 0, 1, 92, 60, False),
        (80, 79, 0, 0, 1, 92, 60, False),
        (80, 45, 40, 0, 1, 92, 60, True),
        (80, 100, 0, 40, 0, 92, 60, True),
        (95.5, 84, 0, 5, 5, 94, 60, True),
    ],
)
def test_grip_error_detection(
    controller: OT3Controller,
    expected_grip_width: float,
    actual_grip_width: float,
    wider: float,
    narrower: float,
    allowed_error: float,
    hard_max: float,
    hard_min: float,
    raise_error: bool,
) -> None:
    context = cast(
        AbstractContextManager[None],
        pytest.raises(FailedGripperPickupError) if raise_error else does_not_raise(),
    )
    with context:
        controller.check_gripper_position_within_bounds(
            expected_grip_width,
            wider,
            narrower,
            actual_grip_width,
            allowed_error,
            hard_min,
            hard_max,
        )


@pytest.mark.parametrize(
    "expected_grip_width,actual_grip_width,wider,narrower,allowed_error,hard_max,hard_min,raise_error",
    [
        (95.5, 84, 0, 5, 5, 94, 60, False),
        (95.5, 60, 0, 5, 5, 94, 60, True),
        (95.5, 94, 0, 5, 5, 94, 60, True),
    ],
)
def test_grip_error_detection_disable_geometry(
    controller: OT3Controller,
    expected_grip_width: float,
    actual_grip_width: float,
    wider: float,
    narrower: float,
    allowed_error: float,
    hard_max: float,
    hard_min: float,
    raise_error: bool,
) -> None:
    context = cast(
        AbstractContextManager[None],
        pytest.raises(FailedGripperPickupError) if raise_error else does_not_raise(),
    )
    with context:
        controller.check_gripper_position_within_bounds(
            expected_grip_width,
            wider,
            narrower,
            actual_grip_width,
            allowed_error,
            hard_min,
            hard_max,
            True,
        )


def move_group_run_side_effect(
    controller: OT3Controller, target_pos: Dict[Axis, float]
) -> Iterator[Dict[NodeId, MotorPositionStatus]]:
    """Return homed position for axis that is present and was commanded to home."""
    motor_nodes = controller._motor_nodes()
    target_nodes = {axis_to_node(ax): ax for ax in target_pos.keys() if ax != Axis.Q}
    res = {}
    for node in motor_nodes:
        pos = 0.0
        if target_nodes.get(node):
            pos = target_pos[target_nodes[node]]
        res[node] = MotorPositionStatus(pos, pos, True, True, MoveCompleteAck(1))
    yield res


@pytest.mark.parametrize(
    argnames=["origin_pos", "target_pos", "expected_pos", "gear_position"],
    argvalues=[
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
                Axis.Q: 0,
            },
            {Axis.Q: 10},
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            10,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            {Axis.Q: 10},
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
            },
            {
                Axis.X: 10,
                Axis.Y: 10,
                Axis.Z_L: 10,
            },
            {
                Axis.X: 10,
                Axis.Y: 10,
                Axis.Z_L: 10,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
            },
            {
                Axis.X: 0,  # Zero Length Move, make sure it doesn't raise an error
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.P_L: 0,
            },
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
    ],
)
async def test_controller_move(
    controller: OT3Controller,
    mock_present_devices: mock.AsyncMock,
    origin_pos: Dict[Axis, float],
    target_pos: Dict[Axis, float],
    expected_pos: Dict[Axis, float],
    gear_position: Optional[float],
) -> None:
    from copy import deepcopy

    controller.update_constraints_for_gantry_load(GantryLoad.HIGH_THROUGHPUT_1000)

    run_target_pos = deepcopy(target_pos)
    config = {"run.side_effect": move_group_run_side_effect(controller, run_target_pos)}
    with mock.patch(  # type: ignore [call-overload]
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner",
        spec=MoveGroupRunner,
        **config,
    ):
        await controller.move(origin_pos, target_pos, 100)
        position = await controller.update_position()
        gear_position = controller.gear_motor_position

        assert position == expected_pos
        assert gear_position == gear_position


@pytest.mark.parametrize(
    argnames=["origin_pos", "target_pos", "expected_pos", "gear_position"],
    argvalues=[
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            {
                Axis.X: 10,
                Axis.Y: 10,
                Axis.Z_L: 50,
                Axis.P_L: 70,
            },
            {
                Axis.X: 10,
                Axis.Y: 10,
                Axis.Z_L: 50,
                Axis.Z_R: 0,
                Axis.P_L: 70,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            {
                Axis.X: 20,
                Axis.Y: 20,
                Axis.Z_L: 10,
                Axis.P_L: 48,
            },
            {
                Axis.X: 20,
                Axis.Y: 20,
                Axis.Z_L: 10,
                Axis.Z_R: 0,
                Axis.P_L: 48,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            {
                Axis.P_L: 70,
            },
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 70,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
        [
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 0,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            {
                Axis.P_L: 30,  # Too short to hit top speed
            },
            {
                Axis.X: 0,
                Axis.Y: 0,
                Axis.Z_L: 0,
                Axis.Z_R: 0,
                Axis.P_L: 30,
                Axis.P_R: 0,
                Axis.Z_G: 0,
                Axis.G: 0,
            },
            None,
        ],
    ],
)
async def test_controller_move_dynamic(
    controller: OT3Controller,
    mock_present_devices: mock.AsyncMock,
    origin_pos: Dict[Axis, float],
    target_pos: Dict[Axis, float],
    expected_pos: Dict[Axis, float],
    gear_position: Optional[float],
) -> None:
    from copy import deepcopy

    controller.update_constraints_for_gantry_load(GantryLoad.LOW_THROUGHPUT)

    run_target_pos = deepcopy(target_pos)
    config = {"run.side_effect": move_group_run_side_effect(controller, run_target_pos)}
    with mock.patch(  # type: ignore [call-overload]
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner",
        spec=MoveGroupRunner,
        **config,
    ):
        await controller.move(origin_pos, target_pos, 70)
        position = await controller.update_position()
        gear_position = controller.gear_motor_position

        assert position == expected_pos
        assert gear_position == gear_position


@pytest.mark.parametrize(
    argnames=["axes", "pipette_has_sensor"],
    argvalues=[[[Axis.P_L, Axis.P_R], True], [[Axis.P_L, Axis.P_R], False]],
)
async def test_pressure_disable(
    controller: OT3Controller,
    axes: List[Axis],
    mock_present_devices: None,
    mock_check_overpressure: None,
    pipette_has_sensor: bool,
) -> None:
    config = {"run.side_effect": move_group_run_side_effect_home(controller, axes)}
    with mock.patch(  # type: ignore [call-overload]
        "opentrons.hardware_control.backends.ot3controller.MoveGroupRunner",
        spec=MoveGroupRunner,
        **config,
    ):
        with mock.patch.object(controller, "_monitor_overpressure") as monitor:
            controller.set_pressure_sensor_available(Axis.P_L, pipette_has_sensor)
            controller.set_pressure_sensor_available(Axis.P_R, True)

            await controller.home(axes, GantryLoad.LOW_THROUGHPUT)

            if pipette_has_sensor:
                monitor.assert_called_once_with(
                    [NodeId.pipette_left, NodeId.pipette_right]
                )
            else:
                monitor.assert_called_once_with([NodeId.pipette_right])



"""Tests for opentrons.protocol_api."""

from typing import List, Optional, overload

from hypothesis import strategies as st

from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION_FOR_FLEX,
)
from opentrons.protocols.api_support.types import APIVersion


def versions_at_or_above(from_version: APIVersion) -> List[APIVersion]:
    """Get a list of versions >= the specified one."""
    return versions_between(
        low_inclusive_bound=from_version, high_inclusive_bound=MAX_SUPPORTED_VERSION
    )


def versions_at_or_below(
    from_version: APIVersion, flex_only: bool = False
) -> List[APIVersion]:
    """Get a list of versions <= the specified one.

    Since there are different minimum supported versions for Flex and OT-2, specify which you care about
    with the second argument.
    """
    if flex_only:
        return versions_between(
            low_inclusive_bound=MIN_SUPPORTED_VERSION_FOR_FLEX,
            high_inclusive_bound=from_version,
        )
    else:
        return versions_between(
            low_inclusive_bound=MIN_SUPPORTED_VERSION, high_inclusive_bound=from_version
        )


def versions_above(from_version: APIVersion) -> List[APIVersion]:
    """Get a list of versions > the specified one."""
    return versions_between(
        low_exclusive_bound=from_version, high_inclusive_bound=MAX_SUPPORTED_VERSION
    )


def versions_below(from_version: APIVersion, flex_only: bool) -> List[APIVersion]:
    """Get a list of versions < the specified one.

    Since there are different minimum supported versions for Flex and OT-2, specify which you care about
    with the second argument.
    """
    if flex_only:
        return versions_between(
            low_inclusive_bound=MIN_SUPPORTED_VERSION_FOR_FLEX,
            high_exclusive_bound=from_version,
        )
    else:
        return versions_between(
            low_inclusive_bound=MIN_SUPPORTED_VERSION, high_exclusive_bound=from_version
        )


@overload
def versions_between(
    *,
    low_inclusive_bound: APIVersion,
    high_inclusive_bound: APIVersion,
) -> List[APIVersion]: ...


@overload
def versions_between(
    *, low_inclusive_bound: APIVersion, high_exclusive_bound: APIVersion
) -> List[APIVersion]: ...


@overload
def versions_between(
    *,
    high_inclusive_bound: APIVersion,
    low_exclusive_bound: APIVersion,
) -> List[APIVersion]: ...


@overload
def versions_between(
    *, low_exclusive_bound: APIVersion, high_exclusive_bound: APIVersion
) -> List[APIVersion]: ...


def versions_between(
    low_inclusive_bound: Optional[APIVersion] = None,
    high_inclusive_bound: Optional[APIVersion] = None,
    low_exclusive_bound: Optional[APIVersion] = None,
    high_exclusive_bound: Optional[APIVersion] = None,
) -> List[APIVersion]:
    """Build a list of versions based on exclusive and inclusive constraints."""
    if low_inclusive_bound and high_inclusive_bound:
        assert low_inclusive_bound.major == high_inclusive_bound.major, (
            "You need to change this test when you add a new major version"
        )
        major = low_inclusive_bound.major
        start = low_inclusive_bound.minor
        stop = high_inclusive_bound.minor + 1
    elif low_inclusive_bound and high_exclusive_bound:
        assert low_inclusive_bound.major == high_exclusive_bound.major, (
            "You need to change this test when you add a new major version"
        )
        major = low_inclusive_bound.major
        start = low_inclusive_bound.minor
        stop = high_exclusive_bound.minor
    elif low_exclusive_bound and high_inclusive_bound:
        assert low_exclusive_bound.major == high_inclusive_bound.major, (
            "You need to change this test when you add a new major version"
        )
        major = low_exclusive_bound.major
        start = low_exclusive_bound.minor + 1
        stop = high_inclusive_bound.minor + 1
    elif low_exclusive_bound and high_exclusive_bound:
        assert low_exclusive_bound.major == high_exclusive_bound.major, (
            "You need to change this test when you add a new major version"
        )
        major = low_exclusive_bound.major
        start = low_exclusive_bound.minor + 1
        stop = high_exclusive_bound.minor
    else:
        raise ValueError("You must specify one low bound and one high bound")
    return [APIVersion(major, minor) for minor in range(start, stop)]


invalid_values = st.one_of(
    st.text(min_size=0, max_size=5),
    st.dictionaries(keys=st.text(), values=st.integers()),
    st.lists(st.integers(min_value=-100, max_value=100), min_size=1, max_size=5),
    st.tuples(st.booleans()),
    st.binary(min_size=1, max_size=5),
)

boolean_looking_values = st.one_of(
    st.just("True"),
    st.just("1"),
    st.just(1),
    st.just("False"),
    st.just("0"),
    st.just(0),
)

reasonable_floats = st.floats(
    min_value=-1000, max_value=1000, allow_infinity=False, allow_nan=False
)
reasonable_ints = st.integers(min_value=-1000, max_value=1000)

reasonable_numbers = st.one_of(reasonable_floats, reasonable_ints)

negative_or_zero_floats = st.floats(
    min_value=-1000, max_value=0, allow_infinity=False, allow_nan=False
)
negative_or_zero_ints = st.integers(min_value=-1000, max_value=0)
negative_or_zero_floats_and_ints = st.one_of(
    negative_or_zero_floats, negative_or_zero_ints
)
positive_or_zero_ints = st.integers(min_value=0, max_value=1000)
positive_or_zero_floats = st.floats(
    min_value=0, max_value=1000, allow_infinity=False, allow_nan=False
)
positive_or_zero_floats_or_ints = st.one_of(
    positive_or_zero_floats, positive_or_zero_ints
)

negative_non_zero_floats = st.floats(
    min_value=-1000, max_value=-0.0001, allow_infinity=False, allow_nan=False
)
negative_non_zero_ints = st.integers(min_value=-1000, max_value=-1)
negative_non_zero_floats_and_ints = st.one_of(
    negative_non_zero_floats, negative_non_zero_ints
)

positive_non_zero_floats = st.floats(
    min_value=0.0001, max_value=1000, allow_infinity=False, allow_nan=False
)
positive_non_zero_ints = st.integers(min_value=1, max_value=1000)
positive_non_zero_floats_and_ints = st.one_of(
    positive_non_zero_floats, positive_non_zero_ints
)



"""Test drop tip in place commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.drop_tip_in_place import (
    DropTipInPlaceImplementation,
    DropTipInPlaceParams,
    DropTipInPlaceResult,
)
from opentrons.protocol_engine.commands.pipetting_common import (
    TipPhysicallyAttachedError,
)
from opentrons.protocol_engine.errors.exceptions import TipAttachedError
from opentrons.protocol_engine.execution import GantryMover, TipHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state.update_types import (
    PipetteTipStateUpdate,
    PipetteUnknownFluidUpdate,
    StateUpdate,
)
from opentrons.types import Point


@pytest.fixture
def mock_tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mock TipHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.fixture
def mock_model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def mock_gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mock GantryMover."""
    return decoy.mock(cls=GantryMover)


async def test_success(
    decoy: Decoy,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
    mock_gantry_mover: GantryMover,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = DropTipInPlaceImplementation(
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
        gantry_mover=mock_gantry_mover,
    )
    params = DropTipInPlaceParams(pipetteId="abc", homeAfter=False)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=DropTipInPlaceResult(),
        state_update=StateUpdate(
            pipette_tip_state=PipetteTipStateUpdate(
                pipette_id="abc", tip_geometry=None, tip_source=None
            ),
            pipette_aspirated_fluid=PipetteUnknownFluidUpdate(pipette_id="abc"),
        ),
    )

    decoy.verify(
        await mock_tip_handler.drop_tip(pipette_id="abc", home_after=False),
        times=1,
    )


async def test_tip_attached_error(
    decoy: Decoy,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
    mock_gantry_mover: GantryMover,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = DropTipInPlaceImplementation(
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
        gantry_mover=mock_gantry_mover,
    )

    params = DropTipInPlaceParams(pipetteId="abc", homeAfter=False)

    decoy.when(await mock_gantry_mover.get_position(pipette_id="abc")).then_return(
        Point(9, 8, 7)
    )
    decoy.when(
        await mock_tip_handler.drop_tip(pipette_id="abc", home_after=False)  # type: ignore[func-returns-value]
    ).then_raise(TipAttachedError("Egads!"))

    decoy.when(mock_model_utils.generate_id()).then_return("error-id")
    decoy.when(mock_model_utils.get_timestamp()).then_return(
        datetime(year=1, month=2, day=3)
    )

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=TipPhysicallyAttachedError.model_construct(
            id="error-id",
            createdAt=datetime(year=1, month=2, day=3),
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (9, 8, 7)},
        ),
        state_update=StateUpdate(
            pipette_aspirated_fluid=PipetteUnknownFluidUpdate(pipette_id="abc")
        ),
        state_update_if_false_positive=StateUpdate(
            pipette_tip_state=PipetteTipStateUpdate(
                pipette_id="abc",
                tip_geometry=None,
                tip_source=None,
            ),
        ),
    )



"""opentrons.system.ffmpeg: Functions and data for interacting with FFMPEG."""

import asyncio
import logging
from typing import Tuple

from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_engine.resources.camera_provider import CameraError

log = logging.getLogger(__name__)

# === FFMPEG Filter Details ===
# The following filters are utilized via the '-vf' flag to manipulate the final image returned:
# 'crop' = [output_width]:[output_height]:x:y
#   The crop is composed of a desired output width and height for the image, and
#   an X/Y position to begin the crop at (becomes the top left of the new image).
# 'scale' = [width]:[height]
#   The resolution of the final image to export, scales up or down based on configuration.
# 'lut' (Look-Up Table) = 'y' (Luminance) = 'val' (Current value of a given pixel from 0-255)
#   The equation on the Look up table takes the current luminance value of an image per pixel
#   and manipulates it using Contrast and Brightness settings. This is applied to the whole image.
# 'hue' (Image color range) = 's' (Saturation) = [range]
#   The hue flag uses the 's' (saturation) modifier to scale image color intensity, default is 1.

# todo(chb, 2025-10-13): Right now we're just zooming towards the center of the frame. The 'pan'
# setting should be used on the latter half of 'crop' to determine our cropping location instead.


async def ffmpeg_capture_image_bytes(
    robot_type: RobotType,
    resolution: Tuple[int, int],
    camera: str,
    zoom: float,
    pan: Tuple[int, int],
    contrast: float,
    brightness: int,
    saturation: float,
) -> bytes | CameraError:
    """Execute an FFMPEG command to capture an image based on various image parameters."""
    if robot_type == "OT-2 Standard":
        ot2_brightness: float = (
            brightness / 128
        ) * -1  # OT-2's equilizer field takes a value of -1.0 to 1.0 for brightness
        command = [
            "ffmpeg",
            "-hwaccel",
            "auto",
            "-video_size",
            f"{resolution[0]}x{resolution[1]}",
            "-f",
            "v4l2",
            "-i",
            f"{camera}",
            "-vf",
            f"crop=iw/{zoom}:ih/{zoom}:(iw-iw/{zoom})/{zoom}:(ih-ih/{zoom})/{zoom},"
            f"scale={resolution[0]}:{resolution[1]},"
            f"eq=brightness={ot2_brightness}:contrast={contrast}:saturation={saturation}",
            "-frames:v",
            "1",
            "-f",
            "image2pipe",
            "-vcodec",
            "mjpeg",
            "-",
        ]
    else:
        command = [
            "ffmpeg",
            "-hwaccel",
            "auto",
            "-video_size",
            f"{resolution[0]}x{resolution[1]}",
            "-f",
            "v4l2",
            "-i",
            f"{camera}",
            "-vf",
            f"crop=iw/{zoom}:ih/{zoom}:(iw-iw/{zoom})/{zoom}:(ih-ih/{zoom})/{zoom},"
            f"scale={resolution[0]}:{resolution[1]},"
            f"lut=y=(val-128)*{contrast}+128-{brightness},"
            f"hue=s={saturation},format=nv12",
            "-frames:v",
            "1",
            "-f",
            "image2pipe",
            "-vcodec",
            "mjpeg",
            "-",
        ]

    subprocess = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout: bytes
    stderr: bytes
    stdout, stderr = await subprocess.communicate()
    if subprocess.returncode == 0:
        log.info("Successfully captured an image with camera.")
        # Upon success, dump our byte stream to the result
        return stdout
    else:
        log.error(
            f"Failed to capture an image with camera, returncode: {subprocess.returncode}, stdout: {stdout.decode()}, stderr: {stderr.decode()}"
        )
        return CameraError(
            message="Failed to return bytes from FFMPEG image capture.",
            code=str(subprocess.returncode)
            if subprocess.returncode is not None
            else None,
        )



"""Common heatershaker base models."""

from opentrons.protocol_engine.state.module_substates.heater_shaker_module_substate import (
    HeaterShakerModuleSubState,
)


async def get_heatershaker_ready_to_shake(
    hs_module_substate: HeaterShakerModuleSubState,
    rpm: float,
) -> int:
    """Check heatershaker state to confirm if it is ready to shake.

    This includes
    - Checking latch closure
    - Validating target speed
    """
    hs_module_substate.raise_if_labware_latch_not_closed()
    # Verify speed from hs module view
    validated_speed = hs_module_substate.validate_target_speed(rpm)
    return validated_speed



"""Test verify tip presence commands."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.verify_tip_presence import (
    VerifyTipPresenceImplementation,
    VerifyTipPresenceParams,
    VerifyTipPresenceResult,
)
from opentrons.protocol_engine.execution import TipHandler
from opentrons.protocol_engine.types import TipPresenceStatus


async def test_verify_tip_presence_implementation(
    decoy: Decoy,
    tip_handler: TipHandler,
) -> None:
    """A VerifyTipPresence command should have an execution implementation."""
    subject = VerifyTipPresenceImplementation(tip_handler=tip_handler)
    data = VerifyTipPresenceParams(
        pipetteId="pipette-id",
        expectedState=TipPresenceStatus.PRESENT,
    )

    result = await subject.execute(data)

    assert result == SuccessData(public=VerifyTipPresenceResult())
    decoy.verify(
        await tip_handler.verify_tip_presence(
            pipette_id="pipette-id",
            expected=TipPresenceStatus.PRESENT,
            follow_singular_sensor=None,
        )
    )



"""Tests for the LabwareStore+LabwareState+LabwareView trifecta.

The trifecta is tested here as a single unit, treating LabwareState as a private
implementation detail.
"""

import pytest

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition3,
    Parameters3,
    RectangularWellDefinition3,
)

from opentrons.protocol_engine import actions, commands, errors
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.labware import (
    LabwareStore,
    LabwareView,
)
from opentrons.protocol_engine.types import DeckSlotLocation, LoadedLabware
from opentrons.types import DeckSlotName


def _dummy_command() -> commands.Command:
    """Return a placeholder command."""
    return commands.Comment.model_construct()  # type: ignore[call-arg]


def test_get_labware(ot3_standard_deck_def: DeckDefinitionV5) -> None:
    """It should get a loaded labware."""
    subject = LabwareStore(deck_definition=ot3_standard_deck_def, deck_fixed_labware=[])

    labware_definition = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        namespace="foo",
        parameters=Parameters3.model_construct(loadName="load_name"),  # type: ignore[call-arg]
        version=123,
    )

    load_labware_update = update_types.LoadedLabwareUpdate(
        labware_id="labware-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        offset_id=None,
        display_name="Display Name",
        definition=labware_definition,
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(loaded_labware=load_labware_update),
            command=_dummy_command(),
        )
    )

    assert LabwareView(subject.state).get(
        "labware-id"
    ) == LoadedLabware.model_construct(
        id="labware-id",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        loadName="load_name",
        definitionUri="foo/load_name/123",
        offsetId=None,
        displayName="Display Name",
    )


def test_raise_if_not_tip_rack(ot3_standard_deck_def: DeckDefinitionV5) -> None:
    """It should raise if the labware is not a tip rack."""
    subject = LabwareStore(deck_definition=ot3_standard_deck_def, deck_fixed_labware=[])
    subject_view = LabwareView(subject.state)

    tiprack_definition = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        namespace="foo",
        parameters=Parameters3.model_construct(loadName="load_name", isTiprack=True),  # type: ignore[call-arg]
        version=123,
    )
    load_tiprack_update = update_types.LoadedLabwareUpdate(
        labware_id="tiprack-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A2),
        offset_id=None,
        display_name="Display Name",
        definition=tiprack_definition,
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(loaded_labware=load_tiprack_update),
            command=_dummy_command(),
        )
    )
    subject_view.raise_if_not_tip_rack("tiprack-id")

    labware_definition = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        namespace="foo",
        parameters=Parameters3.model_construct(loadName="load_name", isTiprack=False),  # type: ignore[call-arg]
        version=123,
    )
    load_labware_update = update_types.LoadedLabwareUpdate(
        labware_id="labware-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        offset_id=None,
        display_name="Display Name",
        definition=labware_definition,
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(loaded_labware=load_labware_update),
            command=_dummy_command(),
        )
    )
    with pytest.raises(errors.LabwareIsNotTipRackError):
        subject_view.raise_if_not_tip_rack("labware-id")


def test_raise_if_wells_are_invalid(ot3_standard_deck_def: DeckDefinitionV5) -> None:
    """It should raise if the wells given do not exist for that labware."""
    subject = LabwareStore(deck_definition=ot3_standard_deck_def, deck_fixed_labware=[])
    subject_view = LabwareView(subject.state)

    labware_definition = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        namespace="foo",
        parameters=Parameters3.model_construct(loadName="load_name"),  # type: ignore[call-arg]
        version=123,
        wells={
            "well-1": RectangularWellDefinition3.model_construct(),  # type: ignore[call-arg]
            "well-2": RectangularWellDefinition3.model_construct(),  # type: ignore[call-arg]
        },
    )
    load_labware_update = update_types.LoadedLabwareUpdate(
        labware_id="labware-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        offset_id=None,
        display_name="Display Name",
        definition=labware_definition,
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(loaded_labware=load_labware_update),
            command=_dummy_command(),
        )
    )

    subject_view.raise_if_wells_are_invalid("labware-id", [])
    subject_view.raise_if_wells_are_invalid("labware-id", ["well-1"])
    subject_view.raise_if_wells_are_invalid("labware-id", ["well-2"])
    subject_view.raise_if_wells_are_invalid("labware-id", ["well-1", "well-2"])

    with pytest.raises(errors.WellDoesNotExistError):
        subject_view.raise_if_wells_are_invalid("labware-id", ["well-1, well-3"])
    with pytest.raises(errors.WellDoesNotExistError):
        subject_view.raise_if_wells_are_invalid(
            "labware-id", ["well-1, well-2", "well-4"]
        )
    with pytest.raises(errors.WellDoesNotExistError):
        subject_view.raise_if_wells_are_invalid("labware-id", ["well-5"])



from datetime import datetime, timezone

import pytest

from opentrons.calibration_storage import deck_configuration as subject
from opentrons.calibration_storage.types import CutoutFixturePlacement


def test_deck_configuration_serdes() -> None:
    """Test that deck configuration serialization/deserialization survives a round trip."""

    dummy_cutout_fixture_placements = [
        CutoutFixturePlacement(
            cutout_fixture_id="a", cutout_id="b", opentrons_module_serial_number="1"
        ),
        CutoutFixturePlacement(
            cutout_fixture_id="c", cutout_id="d", opentrons_module_serial_number="2"
        ),
    ]
    dummy_datetime = datetime(year=1961, month=5, day=6, tzinfo=timezone.utc)

    serialized = subject.serialize_deck_configuration(
        dummy_cutout_fixture_placements, dummy_datetime
    )
    deserialized = subject.deserialize_deck_configuration(serialized)
    assert deserialized == (dummy_cutout_fixture_placements, dummy_datetime)


@pytest.mark.parametrize(
    "input",
    [
        b'{"hello": "world"}',  # Valid JSON, but not valid for the model.
        "😾".encode("utf-8"),  # Not valid JSON.
    ],
)
def test_deserialize_deck_configuration_error_handling(input: bytes) -> None:
    """Test that deserialization handles errors gracefully."""
    assert subject.deserialize_deck_configuration(input) is None



"""Test file for command validations."""

import pytest
from decoy import Decoy

from opentrons.hardware_control.api import API
from opentrons.hardware_control.protocols.types import FlexRobotType, OT2RobotType
from opentrons.protocol_engine.errors.exceptions import HardwareNotSupportedError
from opentrons.protocol_engine.resources.ot3_validation import ensure_ot3_hardware


@pytest.mark.ot3_only
def test_ensure_ot3_hardware(decoy: Decoy) -> None:
    """Should return a OT-3 hardware api."""
    try:
        # TODO (tz, 9-23-22) Figure out a better way to run this test with OT-3 api only.
        from opentrons.hardware_control.ot3api import OT3API

        ot_3_hardware_api = decoy.mock(cls=OT3API)
        decoy.when(ot_3_hardware_api.get_robot_type()).then_return(FlexRobotType)
        result = ensure_ot3_hardware(
            ot_3_hardware_api,
        )
        assert result == ot_3_hardware_api
    except ImportError:
        pass


@pytest.mark.ot3_only
def test_ensure_ot3_hardware_raises_error(decoy: Decoy) -> None:
    """Should raise a HardwareNotSupportedError exception."""
    ot_2_hardware_api = decoy.mock(cls=API)
    decoy.when(ot_2_hardware_api.get_robot_type()).then_return(OT2RobotType)
    with pytest.raises(HardwareNotSupportedError):
        ensure_ot3_hardware(
            ot_2_hardware_api,
        )



"""Tests for opentrons.protocol_reader.file_identifier."""

import json
from dataclasses import dataclass

import pytest
from decoy import Decoy

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_reader.file_identifier import (
    FileIdentificationError,
    FileIdentifier,
    IdentifiedJsonMain,
    IdentifiedLabwareDefinition,
    IdentifiedPythonMain,
)
from opentrons.protocol_reader.file_reader_writer import BufferedFile
from opentrons.protocol_reader.protocol_source import Metadata
from opentrons.protocols import parse
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.types import MalformedPythonProtocolError, PythonProtocol


# TODO(mm, 2023-08-08): Make this autouse=True (apply to all test functions) when
# FileReaderWriter delegates to opentrons.protocols.parse for JSON files.
@pytest.fixture
def use_mock_parse(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Replace opentrons.protocols.parse.parse() with a mock."""
    mock_parse = decoy.mock(func=parse.parse)
    monkeypatch.setattr(parse, "parse", mock_parse)


@pytest.mark.parametrize("filename", ["protocol.py", "protocol.PY", "protocol.Py"])
@pytest.mark.parametrize("python_parse_mode", parse.PythonParseMode)
async def test_python_parsing(
    decoy: Decoy,
    use_mock_parse: None,
    filename: str,
    python_parse_mode: parse.PythonParseMode,
) -> None:
    """It should use opentrons.protocols.parse() to extract basic ID info out of Python files."""
    input_file = BufferedFile(name=filename, contents=b"contents", path=None)

    decoy.when(
        parse.parse(b"contents", filename, python_parse_mode=python_parse_mode)
    ).then_return(
        PythonProtocol(
            api_level=APIVersion(2, 1),
            robot_type="OT-3 Standard",
            metadata={"Hello": "World"},
            text="",
            filename="",
            contents=None,
            bundled_data=None,
            bundled_labware=None,
            bundled_python=None,
            extra_labware=None,
        )
    )

    subject = FileIdentifier()
    [result] = await subject.identify([input_file], python_parse_mode=python_parse_mode)

    assert result == IdentifiedPythonMain(
        original_file=input_file,
        api_level=APIVersion(2, 1),
        robot_type="OT-3 Standard",
        metadata={"Hello": "World"},
    )


@dataclass
class _ValidJsonProtocolSpec:
    file_name: str
    contents: bytes
    expected_schema_version: int
    expected_robot_type: RobotType
    expected_metadata: Metadata


# todo(mm, 2021-09-13): Some of these tests overlap with opentrons.protocols.parse.
# Delegate to opentrons.protocols.parse and test with mocks here.
@pytest.mark.parametrize(
    "spec",
    [
        # Basic JSON protocols of various versions:
        _ValidJsonProtocolSpec(
            file_name="foo.json",
            contents=load_shared_data("protocol/fixtures/7/simpleV7.json"),
            expected_schema_version=7,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "created": 1223131231,
                "tags": ["unitTest"],
            },
        ),
        _ValidJsonProtocolSpec(
            file_name="foo.json",
            contents=load_shared_data("protocol/fixtures/6/simpleV6.json"),
            expected_schema_version=6,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "created": 1223131231,
                "tags": ["unitTest"],
            },
        ),
        _ValidJsonProtocolSpec(
            file_name="foo.json",
            contents=load_shared_data("protocol/fixtures/5/simpleV5.json"),
            expected_schema_version=5,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "category": None,
                "subcategory": None,
                "created": 1223131231,
                "lastModified": None,
                "tags": ["unitTest"],
            },
        ),
        _ValidJsonProtocolSpec(
            file_name="foo.json",
            contents=load_shared_data("protocol/fixtures/4/simpleV4.json"),
            expected_schema_version=4,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "category": None,
                "subcategory": None,
                "created": 1223131231,
                "lastModified": None,
                "tags": ["unitTest"],
            },
        ),
        _ValidJsonProtocolSpec(
            file_name="foo.json",
            contents=load_shared_data("protocol/fixtures/3/simple.json"),
            expected_schema_version=3,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol v3",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "category": None,
                "subcategory": None,
                "created": 1223131231,
                "lastModified": None,
                "tags": ["unitTest"],
            },
        ),
        # JSON with a weirdly capitalized file extension:
        _ValidJsonProtocolSpec(
            file_name="foo.JsOn",
            contents=load_shared_data("protocol/fixtures/3/simple.json"),
            expected_schema_version=3,
            expected_robot_type="OT-2 Standard",
            expected_metadata={
                "protocolName": "Simple test protocol v3",
                "author": "engineering <engineering@opentrons.com>",
                "description": "A short test protocol",
                "category": None,
                "subcategory": None,
                "created": 1223131231,
                "lastModified": None,
                "tags": ["unitTest"],
            },
        ),
    ],
)
async def test_valid_json_protocol(spec: _ValidJsonProtocolSpec) -> None:
    """It should identify the file as a JSON main file and extract basic info."""
    input_file = BufferedFile(name=spec.file_name, contents=spec.contents, path=None)
    expected_result = IdentifiedJsonMain(
        original_file=input_file,
        schema_version=spec.expected_schema_version,
        robot_type=spec.expected_robot_type,
        metadata=spec.expected_metadata,
        unvalidated_json=json.loads(spec.contents),
    )
    subject = FileIdentifier()
    [result] = await subject.identify(
        [input_file], python_parse_mode=parse.PythonParseMode.NORMAL
    )
    assert result == expected_result


@dataclass
class _ValidLabwareDefinitionSpec:
    file_name: str
    contents: bytes


@pytest.mark.parametrize(
    "spec",
    [
        _ValidLabwareDefinitionSpec(
            file_name="foo.json",
            contents=load_shared_data(
                "labware/definitions/2/armadillo_96_wellplate_200ul_pcr_full_skirt/1.json"
            ),
        ),
        _ValidLabwareDefinitionSpec(
            file_name="foo.json",
            contents=load_shared_data(
                "labware/definitions/2/opentrons_96_tiprack_10ul/1.json"
            ),
        ),
    ],
)
async def test_valid_labware_definition(spec: _ValidLabwareDefinitionSpec) -> None:
    """It should identify the file as a labware definition and extract basic info."""
    input_file = BufferedFile(name=spec.file_name, contents=spec.contents, path=None)
    expected_result = IdentifiedLabwareDefinition(
        original_file=input_file, unvalidated_json=json.loads(spec.contents)
    )
    subject = FileIdentifier()
    [result] = await subject.identify(
        [input_file], python_parse_mode=parse.PythonParseMode.NORMAL
    )
    assert result == expected_result


@dataclass
class _InvalidInputSpec:
    file_name: str
    contents: str
    expected_message: str


@pytest.mark.parametrize(
    "spec",
    [
        # Unrecognized file extension:
        _InvalidInputSpec(
            file_name="protocol.python",
            contents="",
            expected_message="protocol.python has an unrecognized file extension.",
        ),
        _InvalidInputSpec(
            file_name="extensionless",
            contents="",
            expected_message="extensionless has an unrecognized file extension.",
        ),
        _InvalidInputSpec(
            file_name="data.dat",
            contents="",
            expected_message="data.dat has an unrecognized file extension.",
        ),
        # .json file that isn't JSON:
        _InvalidInputSpec(
            file_name="foo.json",
            contents="!@#$%",
            expected_message="foo.json is not valid JSON.",
        ),
        # .json file that doesn't conform to any of our protocol or labware schemas:
        _InvalidInputSpec(
            file_name="foo.json",
            contents="{}",
            expected_message="foo.json is not a known Opentrons format.",
        ),
    ],
)
async def test_invalid_input(spec: _InvalidInputSpec) -> None:
    """It should raise errors on invalid input."""
    input_file = BufferedFile(
        name=spec.file_name, contents=spec.contents.encode("utf-8"), path=None
    )
    subject = FileIdentifier()
    with pytest.raises(FileIdentificationError, match=spec.expected_message):
        await subject.identify(
            [input_file],
            python_parse_mode=parse.PythonParseMode.NORMAL,
        )


async def test_invalid_python_api_level(decoy: Decoy, use_mock_parse: None) -> None:
    """It should check the apiLevel and raise if it's not supported."""
    input_file = BufferedFile(name="filename.py", contents=b"contents", path=None)

    decoy.when(
        parse.parse(
            b"contents", "filename.py", python_parse_mode=parse.PythonParseMode.NORMAL
        )
    ).then_return(
        PythonProtocol(
            api_level=APIVersion(999, 999),
            robot_type="OT-3 Standard",
            metadata=None,
            text="",
            filename="",
            contents=None,
            bundled_data=None,
            bundled_labware=None,
            bundled_python=None,
            extra_labware=None,
        )
    )

    subject = FileIdentifier()

    with pytest.raises(FileIdentificationError, match="999.999 is not supported"):
        await subject.identify(
            [input_file], python_parse_mode=parse.PythonParseMode.NORMAL
        )


async def test_malformed_python(decoy: Decoy, use_mock_parse: None) -> None:
    """It should propagate errors that mean the Python file was malformed."""
    input_file = BufferedFile(name="filename.py", contents=b"contents", path=None)

    decoy.when(
        parse.parse(
            b"contents", "filename.py", python_parse_mode=parse.PythonParseMode.NORMAL
        )
    ).then_raise(
        MalformedPythonProtocolError(
            short_message="message 1", long_additional_message="message 2"
        )
    )

    subject = FileIdentifier()

    with pytest.raises(FileIdentificationError) as exc_info:
        await subject.identify(
            [input_file], python_parse_mode=parse.PythonParseMode.NORMAL
        )

    # TODO(mm, 2023-08-8): We probably want to propagate the longer message too, if there is one.
    # Align with the app+UI team about how to do this safely.
    assert str(exc_info.value) == "message 1"



"""Tests for opentrons.protocol_api.core.core_map.LoadedCoreMap."""

import pytest
from decoy import Decoy

from opentrons.protocol_api import Labware, MagneticModuleContext
from opentrons.protocol_api.core.common import LabwareCore, ModuleCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap


def test_get_nothing() -> None:
    """If you ask for nothing, you get nothing."""
    subject = LoadedCoreMap()
    assert subject.get(None) is None


def test_add_labware(decoy: Decoy) -> None:
    """It should be able to add labwares to the map."""
    labware_core = decoy.mock(cls=LabwareCore)
    other_labware_core = decoy.mock(cls=LabwareCore)
    labware = decoy.mock(cls=Labware)

    subject = LoadedCoreMap()
    subject.add(labware_core, labware)

    assert subject.get(labware_core) is labware

    with pytest.raises(KeyError):
        subject.get(other_labware_core)


def test_add_module(decoy: Decoy) -> None:
    """It should be able to add modules to the map."""
    module_core = decoy.mock(cls=ModuleCore)
    other_module_core = decoy.mock(cls=ModuleCore)
    module = decoy.mock(cls=MagneticModuleContext)

    subject = LoadedCoreMap()
    subject.add(module_core, module)

    assert subject.get(module_core) is module

    with pytest.raises(KeyError):
        subject.get(other_module_core)


def test_get_or_add_labware_builds(decoy: Decoy) -> None:
    """It should be able to get-with-default-adder and build the labware if it's missing."""
    labware_core = decoy.mock(cls=LabwareCore)
    labware = decoy.mock(cls=Labware)

    subject = LoadedCoreMap()
    assert subject.get_or_add(labware_core, lambda lw: labware) == labware


def test_get_or_add_labware_caches(decoy: Decoy) -> None:
    """It should be able to get-with-default-adder and return the labware if it's known."""
    labware_core = decoy.mock(cls=LabwareCore)
    labware = decoy.mock(cls=Labware)
    builder = decoy.mock(name="builder")
    subject = LoadedCoreMap()
    subject.add(labware_core, labware)
    assert subject.get_or_add(labware_core, builder) == labware
    decoy.verify(builder(labware_core), times=0)



"""StatusBar handler."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.hardware_control.types import StatusBarState
from opentrons.protocol_engine.execution.status_bar import StatusBarHandler


@pytest.fixture
def subject(
    hardware_api: HardwareControlAPI,
) -> StatusBarHandler:
    """Create a StatusBarHandler with its dependencies mocked out."""
    return StatusBarHandler(hardware_api=hardware_api)


@pytest.mark.parametrize("setting", [StatusBarState.IDLE, StatusBarState.RUN_COMPLETED])
async def test_set_status_bar(
    decoy: Decoy,
    subject: StatusBarHandler,
    hardware_api: OT2HardwareControlAPI,
    setting: StatusBarState,
) -> None:
    """The hardware controller should be called."""
    await subject.set_status_bar(setting)

    decoy.verify(await hardware_api.set_status_bar_state(state=setting), times=1)


@pytest.mark.parametrize(
    argnames=["setting", "should_be_busy"],
    argvalues=[
        [StatusBarState.IDLE, False],
        [StatusBarState.UPDATING, False],
        [StatusBarState.OFF, False],
        [StatusBarState.HARDWARE_ERROR, True],
        [StatusBarState.SOFTWARE_ERROR, True],
        [StatusBarState.RUN_COMPLETED, True],
        [StatusBarState.RUNNING, True],
        [StatusBarState.PAUSED, True],
    ],
)
async def test_check_status_bar_should_not_be_changed(
    decoy: Decoy,
    subject: StatusBarHandler,
    hardware_api: OT2HardwareControlAPI,
    setting: StatusBarState,
    should_be_busy: bool,
) -> None:
    """IDLE, UPDATING, and OFF should be the only states allowed to change."""
    decoy.when(hardware_api.get_status_bar_state()).then_return(setting)

    assert subject.status_bar_should_not_be_changed() == should_be_busy



"""Common flex stacker base models."""

from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import TYPE_CHECKING, Iterator, Literal, Sequence

from typing_extensions import TypedDict

from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ...errors import ErrorOccurrence
from ...state.update_types import StateUpdate
from ...types import (
    OFF_DECK_LOCATION,
    InStackerHopperLocation,
    LabwareLocation,
    LabwareLocationSequence,
    LabwareOffsetLocationSequence,
    LoadedLabware,
    ModuleLocation,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    StackerStoredLabwareGroup,
)
from opentrons.protocol_engine.errors.exceptions import (
    LabwarePoolNotCompatibleWithModuleError,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.execution.equipment import LoadedLabwarePoolData
    from opentrons.protocol_engine.resources import ModelUtils
    from opentrons.protocol_engine.state.module_substates import FlexStackerSubState
    from opentrons.protocol_engine.state.state import StateView


# The stacker cannot dispense labware where there is no gap between the top surface
# of the bottom labware being dispensed, and bottom surface of the top labware.
# This is because the stacker latch, which holds the labware stack, needs enough
# empty space to free the bottom labware, but still hold the top labware once it
# closes.
STACKER_INCOMPATIBLE_LABWARE = set(["opentrons_tough_universal_lid"])


INITIAL_COUNT_DESCRIPTION = dedent(
    """\
        The number of labware that should be initially stored in the stacker. This number will be silently clamped to
        the maximum number of labware that will fit; do not rely on the parameter to know how many labware are in the stacker.

        This field works with the initialStoredLabware field in a complex way.

        The following must be true for initialCount to be valid:
          - It is not specified, and initialStoredLabware is not specified, in which case the stacker will start empty
          - It is not specified, and initialStoredLabware is specified, in which case the contents of the stacker are entirely
            determined by initialStoredLabware.
          - It is specified, and initialStoredLabware is specified, in which case the length of initialStoredLabware must be
            exactly initialCount, and the contents of the stacker will be determined by initialStoredLabware.
        """
)

INITIAL_STORED_LABWARE_DESCRIPTION = dedent(
    """\
        A list of IDs that should be initially stored in the stacker.

        If specified, the first element of the list is the labware on the physical bottom that will be the first labware retrieved.

        This is a complex field. The following must be true for the field to be valid:
        - If this field is specified, then either initialCount must not be specified, or this field must have exactly initalCount elements
        - Each element must contain an id for each corresponding labware details field (i.e. if lidLabware is specified, each element must have
          a lidLabwareId) and must not contain an id for a corresponding labware details field that is not specified (i.e., if adapterLabware
          is not specified, each element must not have an adapterLabwareId).

        The behavior of the command depends on the values of both this field and initialCount.
        - If this field is not specified and initialCount is not specified, the command will create the maximum number of labware objects
          the stacker can hold according to the labware pool specifications.
        - If this field is not specified and initialCount is specified to be 0, the command will create 0 labware objects and the stacker will be empty.
        - If this field is not specified and initialCount is specified to be non-0, the command will create initialCount labware objects of
          each specified labware type (primary, lid, and adapter), with appropriate positions, and arbitrary IDs, loaded into the stacker
        - If this field is specified (and therefore initialCount is not specified or is specified to be the length of this field) then the
          command will create labware objects with the IDs specified in this field and appropriate positions, loaded into the stacker.

        Behavior is also different depending on whether the labware identified by ID in this field exist or not. Either all labware specified
        in this field must exist, or all must not exist.

        Further,
        - If the labware exist, they must be of the same type as identified in the primaryLabware field.
        - If the labware exist and the adapterLabware field is specified, each labware must be currently loaded on a labware of the same kind as
          specified in the adapterLabware field, and that labware must be loaded off-deck
        - If the labware exist and the adapterLabware field is not specified, each labware must be currently loaded off deck directly
        - If the labware exist and the lidLabware field is specified, each labware must currently have a loaded lid of the same kind as specified
          in the lidLabware field
        - If the labware exist and the lidLabware field is not specified, each labware must not currently have a lid
        - If the labware exist, they must have nothing loaded underneath them or above them other than what is mentioned above

        If all the above are true, when this command executes the labware will be immediately moved into InStackerHopper. If any of the above
        are not true, analysis will fail.
        """
)


class FailedLabware(TypedDict, total=False):
    """Holds the labware ID that would have been involved in a failed command."""

    labwareId: str


class FlexStackerStallOrCollisionError(ErrorOccurrence):
    """Returned when the motor driver detects a stall."""

    isDefined: bool = True
    errorType: Literal["flexStackerStallOrCollision"] = "flexStackerStallOrCollision"

    errorCode: str = ErrorCodes.STACKER_STALL_OR_COLLISION_DETECTED.value.code
    detail: str = ErrorCodes.STACKER_STALL_OR_COLLISION_DETECTED.value.detail

    errorInfo: FailedLabware


class FlexStackerShuttleError(ErrorOccurrence):
    """Returned when the Flex Stacker Shuttle is not in the correct location."""

    isDefined: bool = True
    errorType: Literal["flexStackerShuttleMissing"] = "flexStackerShuttleMissing"

    errorCode: str = ErrorCodes.STACKER_SHUTTLE_MISSING.value.code
    detail: str = ErrorCodes.STACKER_SHUTTLE_MISSING.value.detail

    errorInfo: FailedLabware


class FlexStackerHopperError(ErrorOccurrence):
    """Returned when the Flex Stacker hopper labware presence sensor raises an error."""

    isDefined: bool = True
    errorType: Literal["flexStackerHopperLabwareFailed"] = (
        "flexStackerHopperLabwareFailed"
    )

    errorCode: str = ErrorCodes.STACKER_HOPPER_LABWARE_FAILED.value.code
    detail: str = ErrorCodes.STACKER_HOPPER_LABWARE_FAILED.value.detail

    errorInfo: FailedLabware


class FlexStackerLabwareRetrieveError(ErrorOccurrence):
    """Returned when the labware was not able to get to the shuttle."""

    isDefined: bool = True
    errorType: Literal["flexStackerLabwareRetrieveFailed"] = (
        "flexStackerLabwareRetrieveFailed"
    )

    errorCode: str = ErrorCodes.STACKER_SHUTTLE_LABWARE_FAILED.value.code
    detail: str = ErrorCodes.STACKER_SHUTTLE_LABWARE_FAILED.value.detail
    errorInfo: FailedLabware


class FlexStackerLabwareStoreError(ErrorOccurrence):
    """Returned when the labware was not able to get to the shuttle."""

    isDefined: bool = True
    errorType: Literal["flexStackerLabwareStoreFailed"] = (
        "flexStackerLabwareStoreFailed"
    )

    errorCode: str = ErrorCodes.STACKER_SHUTTLE_LABWARE_FAILED.value.code
    detail: str = ErrorCodes.STACKER_SHUTTLE_LABWARE_FAILED.value.detail
    errorInfo: FailedLabware


class FlexStackerShuttleOccupiedError(ErrorOccurrence):
    """Returned when the Flex Stacker Shuttle is occupied when it shouldn't be."""

    isDefined: bool = True
    errorType: Literal["flexStackerShuttleOccupied"] = "flexStackerShuttleOccupied"

    errorCode: str = ErrorCodes.STACKER_SHUTTLE_OCCUPIED.value.code
    detail: str = ErrorCodes.STACKER_SHUTTLE_OCCUPIED.value.detail
    errorInfo: FailedLabware


@dataclass
class _LabwareDefPair:
    definition: LabwareDefinition
    id: str


@dataclass
class _GroupWithDefs:
    primary: _LabwareDefPair
    adapter: _LabwareDefPair | None
    lid: _LabwareDefPair | None


@dataclass
class LabwareWithLocationSequence:
    """Holds a labware id and location."""

    labwareId: str
    locationSequence: LabwareLocationSequence


@dataclass
class GroupWithLocationSequences:
    """Holds labware id and location for group components."""

    primary: LabwareWithLocationSequence
    adapter: LabwareWithLocationSequence | None
    lid: LabwareWithLocationSequence | None


def _labware_location_seq_for_primary(
    group: StackerStoredLabwareGroup,
    base: LabwareLocationSequence,
) -> LabwareWithLocationSequence:
    if group.adapterLabwareId is not None:
        return LabwareWithLocationSequence(
            labwareId=group.primaryLabwareId,
            locationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId=group.adapterLabwareId, lidId=None
                    )
                ]
                + base
            ),
        )
    else:
        return LabwareWithLocationSequence(
            labwareId=group.primaryLabwareId,
            locationSequence=base,
        )


def _labware_location_seq_for_lid(
    group: StackerStoredLabwareGroup,
    base: LabwareLocationSequence,
) -> LabwareWithLocationSequence | None:
    if group.lidLabwareId is None:
        return None
    elif group.adapterLabwareId is None:
        return LabwareWithLocationSequence(
            labwareId=group.lidLabwareId,
            locationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId=group.primaryLabwareId, lidId=group.lidLabwareId
                    )
                ]
                + base
            ),
        )
    else:
        return LabwareWithLocationSequence(
            labwareId=group.lidLabwareId,
            locationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId=group.primaryLabwareId, lidId=group.lidLabwareId
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId=group.adapterLabwareId, lidId=None
                    ),
                ]
                + base
            ),
        )


def _labware_location_seq_for_adapter(
    group: StackerStoredLabwareGroup,
    base: LabwareLocationSequence,
) -> LabwareWithLocationSequence | None:
    if group.adapterLabwareId is None:
        return None
    return LabwareWithLocationSequence(
        labwareId=group.adapterLabwareId, locationSequence=base
    )


def labware_locations_for_group(
    group: StackerStoredLabwareGroup, base: LabwareLocationSequence
) -> GroupWithLocationSequences:
    """Get the labware and location sequences bound together."""
    return GroupWithLocationSequences(
        primary=_labware_location_seq_for_primary(group, base),
        adapter=_labware_location_seq_for_adapter(group, base),
        lid=_labware_location_seq_for_lid(group, base),
    )


def labware_location_base_sequence(
    sample_group: StackerStoredLabwareGroup,
    state_view: StateView,
    default: LabwareLocationSequence,
) -> LabwareLocationSequence:
    """Get the base location sequence for a labware group, including loading the current location if it exists."""
    first = (
        sample_group.adapterLabwareId
        if sample_group.adapterLabwareId is not None
        else sample_group.primaryLabwareId
    )
    if state_view.labware.known(first):
        return state_view.geometry.get_location_sequence(first)
    return default


def primary_location_sequences(
    groups: list[GroupWithLocationSequences],
) -> list[LabwareLocationSequence]:
    """Collate primary location sequences from lists of labware-plus-location."""
    return [primary_location_sequence(group) for group in groups]


def primary_location_sequence(
    group: GroupWithLocationSequences,
) -> LabwareLocationSequence:
    """Get the location sequence for the primary labware."""
    return group.primary.locationSequence


def adapter_location_sequences(
    groups: list[GroupWithLocationSequences],
) -> list[LabwareLocationSequence] | None:
    """Collate adapter location sequences from lists of labware-plus-location."""

    def _yield_adapters(
        groups: list[GroupWithLocationSequences],
    ) -> Iterator[LabwareLocationSequence]:
        for group in groups:
            seq = adapter_location_sequence(group)
            if seq is None:
                continue
            else:
                yield seq

    adapter_seqs = list(_yield_adapters(groups))
    if len(adapter_seqs) != len(groups):
        return None
    return adapter_seqs


def adapter_location_sequences_with_default(
    groups: list[GroupWithLocationSequences], adapter_def: LabwareDefinition | None
) -> list[LabwareLocationSequence] | None:
    """Collate adapter location sequences unless there is no adapter."""
    if adapter_def is None:
        return None
    else:
        return adapter_location_sequences(groups)


def adapter_location_sequence(
    group: GroupWithLocationSequences,
) -> LabwareLocationSequence | None:
    """Get the adapter location sequence from a group."""
    if group.adapter is None:
        return None
    return group.adapter.locationSequence


def lid_location_sequences(
    groups: list[GroupWithLocationSequences],
) -> list[LabwareLocationSequence] | None:
    """Collate lid location sequences from lists of labware-plus-location."""

    def _yield_lids(
        groups: list[GroupWithLocationSequences],
    ) -> Iterator[LabwareLocationSequence]:
        for group in groups:
            seq = lid_location_sequence(group)
            if seq is None:
                continue
            else:
                yield seq

    lid_seqs = list(_yield_lids(groups))
    if len(lid_seqs) != len(groups):
        return None
    return lid_seqs


def lid_location_sequences_with_default(
    groups: list[GroupWithLocationSequences], lid_def: LabwareDefinition | None
) -> list[LabwareLocationSequence] | None:
    """Collate lid location sequences unless there is no lid."""
    if lid_def is None:
        return None
    else:
        return lid_location_sequences(groups)


def lid_location_sequence(
    group: GroupWithLocationSequences,
) -> LabwareLocationSequence | None:
    """Get the lid location sequence from a group."""
    if group.lid is None:
        return None
    return group.lid.locationSequence


def _check_one_preloaded_labware_known(
    group: StackerStoredLabwareGroup, state_view: StateView
) -> bool:
    """Slightly tricky way to check if a labware group represents known labware.

    Return true if all specified labware are known; false if all specified labware are not known; and
    raise CommandPreconditionViolated if some specified labware are known and some are not.
    """
    if state_view.labware.known(group.primaryLabwareId):
        if group.lidLabwareId is not None and not state_view.labware.known(
            group.lidLabwareId
        ):
            raise CommandPreconditionViolated(
                "Either all labware ids must be known or none must be, but primary and lid do not match"
            )
        if group.adapterLabwareId is not None and not state_view.labware.known(
            group.adapterLabwareId
        ):
            raise CommandPreconditionViolated(
                "Either all labware ids must be known or none must be, but primary and adapter do not match"
            )

        return True
    else:
        if group.lidLabwareId is not None and state_view.labware.known(
            group.lidLabwareId
        ):
            raise CommandPreconditionViolated(
                "Either all labware ids must be known or none must be, but primary and lid do not match"
            )
        if group.adapterLabwareId is not None and state_view.labware.known(
            group.adapterLabwareId
        ):
            raise CommandPreconditionViolated(
                "Either all labware ids must be known or none must be, but primary and lid do not match"
            )
        return False


def _check_one_preloaded_labware(  # noqa: C901
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
    group: StackerStoredLabwareGroup,
    state_view: StateView,
) -> None:
    """Do some preflight checks for labware known to be preloaded.

    Check that labware known to the engine is located appropriately to be loaded to the stacker.
    hopper directly (i.e. during setStoredLabware or fill, NOT with store). While we're at it, bind the def and the id.
    """
    pool_primary_uri = state_view.labware.get_uri_from_definition(
        pool_primary_definition
    )
    stored_primary_uri = state_view.labware.get_definition_uri(group.primaryLabwareId)

    if pool_primary_uri != stored_primary_uri:
        raise CommandPreconditionViolated(
            f"Each labware group must be composed of the same kinds of labware, but previous labware groups specify primary URI {stored_primary_uri} and this one specifies {pool_primary_uri}."
        )
    if pool_adapter_definition:
        if group.adapterLabwareId is None:
            raise CommandPreconditionViolated(
                "Each labware group must be composed of the same kinds of labware, but previous labware groups specify an adapter and this one does not."
            )
        stored_adapter_uri = state_view.labware.get_definition_uri(
            group.adapterLabwareId
        )
        pool_adapter_uri = state_view.labware.get_uri_from_definition(
            pool_adapter_definition
        )
        if stored_adapter_uri != pool_adapter_uri:
            raise CommandPreconditionViolated(
                f"Each labware group must be composed of the same kinds of labware, but previous labware groups specify adapter URI {stored_adapter_uri} and this one specifies {pool_adapter_uri}."
            )
        if state_view.labware.get_location(group.adapterLabwareId) != OFF_DECK_LOCATION:
            raise CommandPreconditionViolated(
                "All existing adapters to be loaded into a stacker must be currently OFF_DECK"
            )
        if state_view.labware.get_location(group.primaryLabwareId) != OnLabwareLocation(
            labwareId=group.adapterLabwareId
        ):
            raise CommandPreconditionViolated(
                "Existing labware groups to be loaded into a stacker must already be associated"
            )
    else:
        if group.adapterLabwareId is not None:
            raise CommandPreconditionViolated(
                "Each labware group must be composed of the same kinds of labware, but previous labware groups specify no adapter and this one does."
            )
        if state_view.labware.get_location(group.primaryLabwareId) != OFF_DECK_LOCATION:
            raise CommandPreconditionViolated(
                "All existing labware without adapters to be loaded into a stacker must be currently OFF_DECK"
            )
    if pool_lid_definition:
        if group.lidLabwareId is None:
            raise CommandPreconditionViolated(
                "Each labware group must be composed of the same kinds of labware, but previous labware groups specify a lid and this one does not."
            )
        stored_lid_uri = state_view.labware.get_definition_uri(group.lidLabwareId)
        pool_lid_uri = state_view.labware.get_uri_from_definition(pool_lid_definition)
        if stored_lid_uri != pool_lid_uri:
            raise CommandPreconditionViolated(
                f"Each labware group must be composed of the same kinds of labware, but previous labware groups specify lid URI {stored_lid_uri} and this one specifies {pool_lid_uri}."
            )

        if (
            state_view.labware.get_location(group.lidLabwareId)
            != OnLabwareLocation(labwareId=group.primaryLabwareId)
            or state_view.labware.get_lid_id_by_labware_id(group.primaryLabwareId)
            != group.lidLabwareId
        ):
            raise CommandPreconditionViolated(
                "Existing labware groups to be loaded into a stacker must already be associated"
            )
    else:
        if group.lidLabwareId is not None:
            raise CommandPreconditionViolated(
                "Each labware group must be composed of the same kinds of labware, but previous labware groups did not specify a lid and this one does."
            )


def check_preloaded_labware(
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
    ids: list[StackerStoredLabwareGroup],
    state_view: StateView,
) -> None:
    """Check whether a list of known-to-be-preloaded labware match the pool constraints."""
    for group in ids:
        _check_one_preloaded_labware(
            pool_primary_definition,
            pool_adapter_definition,
            pool_lid_definition,
            group,
            state_view,
        )


def check_if_labware_preloaded(
    ids: list[StackerStoredLabwareGroup], state_view: StateView
) -> bool:
    """Determine whether the list of ids has already been loaded or needs to be loaded."""
    if len(ids) == 0:
        return False
    first = _check_one_preloaded_labware_known(ids[0], state_view)
    for group in ids[1:]:
        if _check_one_preloaded_labware_known(group, state_view) != first:
            raise CommandPreconditionViolated(
                "All labware must be previously loaded or none must be previously loaded."
            )
    return first


def _add_labware_details_to_dicts(
    definitions_by_id: dict[str, LabwareDefinition],
    display_names_by_id: dict[str, str | None],
    new_locations_by_id: dict[str, LabwareLocation],
    offset_ids_by_id: dict[str, str | None],
    definition: LabwareDefinition,
    labware: LoadedLabware,
) -> None:
    definitions_by_id[labware.id] = definition
    display_names_by_id[labware.id] = None
    new_locations_by_id[labware.id] = labware.location
    offset_ids_by_id[labware.id] = labware.offsetId


def _add_pool_labware_details_to_dicts(
    definitions_by_id: dict[str, LabwareDefinition],
    display_names_by_id: dict[str, str | None],
    new_locations_by_id: dict[str, LabwareLocation],
    offset_ids_by_id: dict[str, str | None],
    lid_parent_ids: list[str],
    lid_ids: list[str],
    pool_primary_definition: LabwareDefinition,
    pool_lid_definition: LabwareDefinition | None,
    pool_adapter_definition: LabwareDefinition | None,
    pool_group: LoadedLabwarePoolData,
) -> None:
    if pool_group.lid_labware:
        assert pool_lid_definition  # safe: only have lids if the pool has lids
        lid_parent_ids.append(pool_group.primary_labware.id)
        lid_ids.append(pool_group.lid_labware.id)
        _add_labware_details_to_dicts(
            definitions_by_id,
            display_names_by_id,
            new_locations_by_id,
            offset_ids_by_id,
            pool_lid_definition,
            pool_group.lid_labware,
        )
    if pool_group.adapter_labware:
        assert pool_adapter_definition
        _add_labware_details_to_dicts(
            definitions_by_id,
            display_names_by_id,
            new_locations_by_id,
            offset_ids_by_id,
            pool_adapter_definition,
            pool_group.adapter_labware,
        )

    _add_labware_details_to_dicts(
        definitions_by_id,
        display_names_by_id,
        new_locations_by_id,
        offset_ids_by_id,
        pool_primary_definition,
        pool_group.primary_labware,
    )


async def build_n_labware_with_ids(
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
    module_id: str,
    ids: list[StackerStoredLabwareGroup],
    current_contained_labware: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
) -> tuple[StateUpdate, list[StackerStoredLabwareGroup]]:
    """Create labware objects to be stored inside the hopper."""
    pool_groups = [
        await equipment.load_labware_pool_from_definitions(
            pool_primary_definition=pool_primary_definition,
            pool_adapter_definition=pool_adapter_definition,
            pool_lid_definition=pool_lid_definition,
            location=InStackerHopperLocation(moduleId=module_id),
            primary_id=id_group.primaryLabwareId,
            adapter_id=id_group.adapterLabwareId,
            lid_id=id_group.lidLabwareId,
        )
        for id_group in ids
    ]
    definitions_by_id: dict[str, LabwareDefinition] = {}
    display_names_by_id: dict[str, str | None] = {}
    new_locations_by_id: dict[str, LabwareLocation] = {}
    offset_ids_by_id: dict[str, str | None] = {}
    lid_parent_ids: list[str] = []
    lid_ids: list[str] = []
    for pool_group in pool_groups:
        _add_pool_labware_details_to_dicts(
            definitions_by_id,
            display_names_by_id,
            new_locations_by_id,
            offset_ids_by_id,
            lid_parent_ids,
            lid_ids,
            pool_primary_definition,
            pool_lid_definition,
            pool_adapter_definition,
            pool_group,
        )
    new_contained_labware = current_contained_labware + ids
    return (
        StateUpdate()
        .update_flex_stacker_contained_labware(module_id, new_contained_labware)
        .set_batch_loaded_labware(
            definitions_by_id=definitions_by_id,
            offset_ids_by_id=offset_ids_by_id,
            display_names_by_id=display_names_by_id,
            new_locations_by_id=new_locations_by_id,
        )
        .set_lids(parent_labware_ids=lid_parent_ids, lid_ids=lid_ids)
    ), new_contained_labware


async def assign_n_labware(
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
    module_id: str,
    ids: list[StackerStoredLabwareGroup],
    current_contained_labware: list[StackerStoredLabwareGroup],
    state_view: StateView,
) -> tuple[StateUpdate, list[StackerStoredLabwareGroup]]:
    """Assign a list of labware to be inside the stacker hopper."""
    check_preloaded_labware(
        pool_primary_definition,
        pool_adapter_definition,
        pool_lid_definition,
        ids,
        state_view,
    )

    def _bottom_labware(group: StackerStoredLabwareGroup) -> str:
        if group.adapterLabwareId:
            return group.adapterLabwareId
        return group.primaryLabwareId

    def _add_ids(
        group: StackerStoredLabwareGroup, offset_dict: dict[str, str | None]
    ) -> None:
        offset_dict[group.primaryLabwareId] = None
        if group.adapterLabwareId:
            offset_dict[group.adapterLabwareId] = None
        if group.lidLabwareId:
            offset_dict[group.lidLabwareId] = None

    new_locations_by_id = {
        _bottom_labware(group): InStackerHopperLocation(moduleId=module_id)
        for group in ids
    }
    new_offset_ids_by_id: dict[str, str | None] = {}
    for group in ids:
        _add_ids(group, new_offset_ids_by_id)

    new_contained_labware = current_contained_labware + ids
    return (
        StateUpdate()
        .update_flex_stacker_contained_labware(module_id, new_contained_labware)
        .set_batch_labware_location(
            new_locations_by_id=new_locations_by_id,
            new_offset_ids_by_id=new_offset_ids_by_id,
        )
    ), new_contained_labware


async def build_or_assign_labware_to_hopper(
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
    module_id: str,
    ids: list[StackerStoredLabwareGroup],
    current_contained_labware: list[StackerStoredLabwareGroup],
    equipment: EquipmentHandler,
    state_view: StateView,
) -> tuple[StateUpdate, list[StackerStoredLabwareGroup]]:
    """Use the common params to labware-creating stacker commands to load labware appropriately.

    If the specified labware IDs exist already, labware is moved; if they don't, labware is created.
    """
    if check_if_labware_preloaded(ids, state_view):
        return await assign_n_labware(
            pool_primary_definition,
            pool_adapter_definition,
            pool_lid_definition,
            module_id,
            ids,
            current_contained_labware,
            state_view,
        )
    else:
        return await build_n_labware_with_ids(
            pool_primary_definition,
            pool_adapter_definition,
            pool_lid_definition,
            module_id,
            ids,
            current_contained_labware,
            equipment,
        )


def _count_from_lw_list_or_initial_count(
    initial_count: int | None,
    initial_lw: list[StackerStoredLabwareGroup] | None,
    max_pool_count: int,
    current_count: int,
) -> int:
    """Count the number of labware to be added to the stacker."""
    capacity = max(max_pool_count - current_count, 0)

    if initial_count is not None:
        if initial_lw and len(initial_lw) != initial_count:
            raise CommandPreconditionViolated(
                "If initialCount and initialStoredLabware are both specified, the number of labware must equal the count"
            )
        to_store_count = initial_count
    elif initial_lw is not None:
        to_store_count = len(initial_lw)
    else:
        # neither initialCount nor initialStoredLabware are specified
        if not capacity:
            raise CommandPreconditionViolated(
                "No labware groups were specified to be stored, but the stacker is already full"
            )
        return capacity

    if to_store_count > capacity:
        error_text = f" and is already holding {current_count}" if current_count else ""
        raise CommandPreconditionViolated(
            f"{to_store_count} labware groups were requested to be stored, "
            f"but the stacker can hold only {max_pool_count}{error_text}"
        )
    return to_store_count


def _build_one_labware_group(
    has_adapter: bool,
    has_lid: bool,
    group: StackerStoredLabwareGroup | None,
    model_utils: ModelUtils,
) -> StackerStoredLabwareGroup:
    if group:
        return group
    return StackerStoredLabwareGroup(
        primaryLabwareId=model_utils.generate_id(),
        adapterLabwareId=(model_utils.generate_id() if has_adapter else None),
        lidLabwareId=(model_utils.generate_id() if has_lid else None),
    )


def build_ids_to_fill(
    has_adapter: bool,
    has_lid: bool,
    initialLabware: list[StackerStoredLabwareGroup] | None,
    initialCount: int | None,
    max_count: int,
    current_count: int,
    model_utils: ModelUtils,
) -> list[StackerStoredLabwareGroup]:
    """Handle the common params for filling the stacker to make a list of ids.

    Only builds labware to add to the current stored (defined by current count).
    """
    count = _count_from_lw_list_or_initial_count(
        initialCount, initialLabware, max_count, current_count
    )

    def _pad_labware_to_count(
        labware_list: list[StackerStoredLabwareGroup], count: int
    ) -> Sequence[StackerStoredLabwareGroup | None]:
        if len(labware_list) < count:
            return labware_list + ([None] * (count - len(labware_list)))
        else:
            return labware_list[:count]

    return [
        _build_one_labware_group(has_adapter, has_lid, group, model_utils)
        for group in _pad_labware_to_count(initialLabware or [], count)
    ]


def build_retrieve_labware_move_updates(
    group: StackerStoredLabwareGroup,
    stacker: FlexStackerSubState,
    state_view: StateView,
) -> tuple[dict[str, LabwareLocation], dict[str, str | None]]:
    """Build the arguments required for batch_labware_location."""
    locations_for_ids: dict[str, LabwareLocation] = {}
    offset_ids_by_id: dict[str, str | None] = {}
    base_offset_location = state_view.geometry.get_projected_offset_location(
        ModuleLocation(moduleId=stacker.module_id)
    )
    assert stacker.pool_primary_definition, "Undefined labware pool"
    primary_uri = state_view.labware.get_uri_from_definition(
        stacker.pool_primary_definition
    )

    def _prepend_loc(
        new: LabwareOffsetLocationSequence,
        current: LabwareOffsetLocationSequence | None,
    ) -> LabwareOffsetLocationSequence | None:
        if current is None:
            return None
        return new + current

    def _find_offset_id(
        uri: str, offset_location: LabwareOffsetLocationSequence | None
    ) -> str | None:
        if offset_location is None:
            return None
        offset = state_view.labware.find_applicable_labware_offset(uri, offset_location)
        if offset is None:
            return None
        return offset.id

    if group.adapterLabwareId:
        locations_for_ids[group.adapterLabwareId] = ModuleLocation(
            moduleId=stacker.module_id
        )
        locations_for_ids[group.primaryLabwareId] = OnLabwareLocation(
            labwareId=group.adapterLabwareId
        )
        assert stacker.pool_adapter_definition, (
            "Mismatched pool and labware definitions"
        )
        adapter_uri = state_view.labware.get_uri_from_definition(
            stacker.pool_adapter_definition
        )
        offset_ids_by_id[group.adapterLabwareId] = _find_offset_id(
            adapter_uri, base_offset_location
        )
        primary_offset_location = _prepend_loc(
            [OnLabwareOffsetLocationSequenceComponent(labwareUri=adapter_uri)],
            base_offset_location,
        )
        offset_ids_by_id[group.primaryLabwareId] = _find_offset_id(
            primary_uri, primary_offset_location
        )

    else:
        locations_for_ids[group.primaryLabwareId] = ModuleLocation(
            moduleId=stacker.module_id
        )
        primary_offset_location = base_offset_location
        offset_ids_by_id[group.primaryLabwareId] = _find_offset_id(
            primary_uri, primary_offset_location
        )

    if group.lidLabwareId:
        assert stacker.pool_lid_definition is not None, (
            "Mismatched pool and stored labware"
        )
        lid_offset_location = _prepend_loc(
            [OnLabwareOffsetLocationSequenceComponent(labwareUri=primary_uri)],
            primary_offset_location,
        )
        locations_for_ids[group.lidLabwareId] = OnLabwareLocation(
            labwareId=group.primaryLabwareId
        )
        offset_ids_by_id[group.lidLabwareId] = _find_offset_id(
            state_view.labware.get_uri_from_definition(stacker.pool_lid_definition),
            lid_offset_location,
        )
    return locations_for_ids, offset_ids_by_id


def validate_labware_pool_compatible_with_stacker(
    pool_primary_definition: LabwareDefinition,
    pool_adapter_definition: LabwareDefinition | None,
    pool_lid_definition: LabwareDefinition | None,
) -> None:
    """Verifies that the given labware pool is compatible with the stacker."""
    labware_pool = set(
        lw.parameters.loadName
        for lw in [
            pool_primary_definition,
            pool_adapter_definition,
            pool_lid_definition,
        ]
        if lw is not None
    )
    incompatible_labware = list(labware_pool & STACKER_INCOMPATIBLE_LABWARE)
    if incompatible_labware:
        raise LabwarePoolNotCompatibleWithModuleError(
            f"The stacker cannot store {incompatible_labware}"
        )



"""Tests for adjacent_slots_getters."""

from typing import List, Optional

import pytest

from opentrons_shared_data.robot.types import RobotType

from opentrons.motion_planning.adjacent_slots_getters import (
    _MixedTypeSlots,
    get_adjacent_slots,
    get_adjacent_staging_slot,
    get_east_slot,
    get_east_west_slots,
    get_north_slot,
    get_north_south_slots,
    get_south_slot,
    get_surrounding_slots,
    get_west_of_staging_slot,
    get_west_slot,
)
from opentrons.types import DeckSlotName, StagingSlotName


@pytest.mark.parametrize(
    argnames=["slot", "expected_east", "expected_west", "expected_both"],
    argvalues=[
        [1, 2, None, [2]],
        [2, 3, 1, [3, 1]],
        [3, None, 2, [2]],
        [4, 5, None, [5]],
        [5, 6, 4, [6, 4]],
        [6, None, 5, [5]],
        [7, 8, None, [8]],
        [8, 9, 7, [9, 7]],
        [9, None, 8, [8]],
        [10, 11, None, [11]],
        [11, 12, 10, [12, 10]],
        [12, None, 11, [11]],
    ],
)
def test_get_east_west_slots(
    slot: int,
    expected_west: Optional[int],
    expected_east: Optional[int],
    expected_both: List[int],
) -> None:
    """It should return a list of slots on the east & west."""
    assert get_east_slot(slot) == expected_east
    assert get_west_slot(slot) == expected_west
    assert get_east_west_slots(slot) == expected_both


@pytest.mark.parametrize(
    argnames=["slot", "expected_north", "expected_south", "expected_both"],
    argvalues=[
        [1, 4, None, [4]],
        [2, 5, None, [5]],
        [3, 6, None, [6]],
        [4, 7, 1, [7, 1]],
        [5, 8, 2, [8, 2]],
        [6, 9, 3, [9, 3]],
        [7, 10, 4, [10, 4]],
        [8, 11, 5, [11, 5]],
        [9, 12, 6, [12, 6]],
        [10, None, 7, [7]],
        [11, None, 8, [8]],
        [12, None, 9, [9]],
    ],
)
def test_get_north_south_slots(
    slot: int,
    expected_north: Optional[int],
    expected_south: Optional[int],
    expected_both: List[int],
) -> None:
    """It should return a list of slots in the north & south."""
    assert get_north_slot(slot) == expected_north
    assert get_south_slot(slot) == expected_south
    assert get_north_south_slots(slot) == expected_both


@pytest.mark.parametrize(
    argnames=["slot", "expected_adjacent"],
    argvalues=[
        [1, [2, 4]],
        [2, [5, 1, 3]],
        [3, [6, 2]],
        [4, [7, 1, 5]],
        [5, [2, 8, 4, 6]],
        [6, [9, 3, 5]],
        [7, [10, 4, 8]],
        [8, [11, 5, 7, 9]],
        [9, [12, 6, 8]],
        [10, [7, 11]],
        [11, [8, 10, 12]],
        [12, [9, 11]],
    ],
)
def test_get_adjacent_slots(slot: int, expected_adjacent: List[int]) -> None:
    """It should return a list of adjacent slots."""
    assert sorted(get_adjacent_slots(slot)) == sorted(expected_adjacent)


@pytest.mark.parametrize(
    argnames=["slot", "expected_adjacent"],
    argvalues=[
        (StagingSlotName.SLOT_A4, DeckSlotName.SLOT_A3),
        (StagingSlotName.SLOT_B4, DeckSlotName.SLOT_B3),
        (StagingSlotName.SLOT_C4, DeckSlotName.SLOT_C3),
        (StagingSlotName.SLOT_D4, DeckSlotName.SLOT_D3),
    ],
)
def test_get_west_of_staging_slot(
    slot: StagingSlotName, expected_adjacent: DeckSlotName
) -> None:
    """It should find the slot directly west of a staging slot."""
    assert get_west_of_staging_slot(slot) == expected_adjacent


@pytest.mark.parametrize(
    argnames=["slot", "expected_adjacent"],
    argvalues=[
        (DeckSlotName.SLOT_A3, StagingSlotName.SLOT_A4),
        (DeckSlotName.SLOT_B3, StagingSlotName.SLOT_B4),
        (DeckSlotName.SLOT_C3, StagingSlotName.SLOT_C4),
        (DeckSlotName.SLOT_D3, StagingSlotName.SLOT_D4),
        (DeckSlotName.SLOT_D1, None),
        (DeckSlotName.SLOT_1, None),
    ],
)
def test_get_adjacent_staging_slot(
    slot: DeckSlotName, expected_adjacent: Optional[StagingSlotName]
) -> None:
    """It should find the adjacent slot east of a staging slot if it exists."""
    assert get_adjacent_staging_slot(slot) == expected_adjacent


@pytest.mark.parametrize(
    argnames=["slot", "robot_type", "expected_surrounding_slots"],
    argvalues=[
        (
            2,
            "OT-2 Standard",
            _MixedTypeSlots(
                regular_slots=[
                    DeckSlotName.SLOT_3,
                    DeckSlotName.SLOT_1,
                    DeckSlotName.SLOT_5,
                    DeckSlotName.SLOT_6,
                    DeckSlotName.SLOT_4,
                ],
                staging_slots=[],
            ),
        ),
        (
            6,
            "OT-2 Standard",
            _MixedTypeSlots(
                regular_slots=[
                    DeckSlotName.SLOT_5,
                    DeckSlotName.SLOT_9,
                    DeckSlotName.SLOT_3,
                    DeckSlotName.SLOT_8,
                    DeckSlotName.SLOT_2,
                ],
                staging_slots=[],
            ),
        ),
        (
            6,
            "OT-3 Standard",
            _MixedTypeSlots(
                regular_slots=[
                    DeckSlotName.SLOT_C2,
                    DeckSlotName.SLOT_B3,
                    DeckSlotName.SLOT_D3,
                    DeckSlotName.SLOT_B2,
                    DeckSlotName.SLOT_D2,
                ],
                staging_slots=[
                    StagingSlotName.SLOT_B4,
                    StagingSlotName.SLOT_C4,
                    StagingSlotName.SLOT_D4,
                ],
            ),
        ),
        (
            10,
            "OT-3 Standard",
            _MixedTypeSlots(
                regular_slots=[
                    DeckSlotName.SLOT_A2,
                    DeckSlotName.SLOT_B1,
                    DeckSlotName.SLOT_B2,
                ],
                staging_slots=[],
            ),
        ),
    ],
)
def test_get_surrounding_slots(
    slot: int,
    robot_type: RobotType,
    expected_surrounding_slots: _MixedTypeSlots,
) -> None:
    """It should get the list of surrounding slots appropriate for the robot type."""
    assert (
        get_surrounding_slots(slot=slot, robot_type=robot_type)
        == expected_surrounding_slots
    )



from typing import AsyncGenerator

import pytest

from .build_module import build_module
from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.emulation.util import TEMPERATURE_ROOM
from opentrons.hardware_control.modules import HeaterShaker

TEMP_ROOM_LOW = TEMPERATURE_ROOM - 0.7
TEMP_ROOM_HIGH = TEMPERATURE_ROOM + 0.7


@pytest.fixture
async def heatershaker(
    emulation_app: None,
    emulator_settings: Settings,
    execution_manager: ExecutionManager,
    poll_interval_seconds: float,
) -> AsyncGenerator[HeaterShaker, None]:
    module = await build_module(
        HeaterShaker,
        port=emulator_settings.heatershaker_proxy.driver_port,
        execution_manager=execution_manager,
        poll_interval_seconds=poll_interval_seconds,
    )
    try:
        yield module
    finally:
        await module.cleanup()


def test_device_info(heatershaker: HeaterShaker) -> None:
    """Confirm device_info returns correct values."""
    assert heatershaker.device_info == {
        "model": "v01",
        "version": "v0.0.1",
        "serial": "heater_shaker_emulator",
    }


async def test_latch_status(heatershaker: HeaterShaker) -> None:
    """It should run open and close latch."""
    await heatershaker._poller.wait_next_poll()
    assert heatershaker.labware_latch_status.value == "idle_open"

    await heatershaker.close_labware_latch()
    assert heatershaker.labware_latch_status.value == "idle_closed"

    await heatershaker.open_labware_latch()
    assert heatershaker.labware_latch_status.value == "idle_open"


async def test_speed(heatershaker: HeaterShaker) -> None:
    """It should speed up, then slow down."""

    await heatershaker.set_speed(550)
    assert heatershaker.target_speed == 550

    # The acceptable delta for actual speed is 100
    assert 450 <= heatershaker.speed <= 650


async def test_deactivate_shaker(heatershaker: HeaterShaker) -> None:
    """It should speed up, then slow down."""

    await heatershaker.set_speed(150)
    assert heatershaker.target_speed == 150

    await heatershaker.deactivate_shaker()

    assert heatershaker.speed == 0
    assert heatershaker.target_speed is None


async def test_deactivate_heater(heatershaker: HeaterShaker) -> None:
    await heatershaker.start_set_temperature(30.0)
    await heatershaker.await_temperature(30.0)
    assert heatershaker.target_temperature == 30.0
    assert 29.3 <= heatershaker.temperature <= 30.7

    await heatershaker.deactivate_heater()
    assert heatershaker.target_temperature is None
    assert TEMP_ROOM_LOW <= heatershaker.temperature <= TEMP_ROOM_HIGH


async def test_temp(heatershaker: HeaterShaker) -> None:
    """Test setting temp"""

    await heatershaker.start_set_temperature(50.0)
    assert heatershaker.target_temperature == 50.0
    assert heatershaker.temperature != 50.0

    await heatershaker.await_temperature(50.0)
    assert heatershaker.target_temperature == 50.0

    # Acceptable delta is 0.7 degrees
    assert 49.3 <= heatershaker.temperature <= 50.7


async def test_forcible_deactivate(
    heatershaker: HeaterShaker, execution_manager: ExecutionManager
) -> None:
    """Can override wait_for_is_running."""
    await execution_manager.pause()
    await heatershaker.deactivate(must_be_running=False)
    await heatershaker.deactivate_heater(must_be_running=False)
    await heatershaker.deactivate_shaker(must_be_running=False)
